import logging
import random
import wandb
import datasets
from pathlib import Path
from datasets import DatasetDict
from transformers import (
    AutoTokenizer,
    AutoModelForCausalLM,
    DataCollatorForLanguageModeling,
    Trainer,
    TrainingArguments,
)
from ekorpkit import eKonf
from tqdm.auto import tqdm
from functools import partialmethod
from ekorpkit.batch import BaseConfig
from .stable import StableDiffusion


def disable_tqdm():
    # THIS IS A DIRTY HACK TO SILENCE THE PROGRESS BAR
    # THE TQDM PROGRESS BAR BOTTLENECKS CELL OUTPUT AND SLOWS THE NOTEBOOK
    tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)


log = logging.getLogger(__name__)


class PromptGenerator(BaseConfig):
    start_token = "<BOP>"
    pad_token = "<PAD>"
    end_token = "<EOP>"

    def __init__(self, root_dir=None, config_name="default", **args):
        cfg = eKonf.compose(f"model/prompt={config_name}")
        cfg = eKonf.merge(cfg, args)
        super().__init__(root_dir=root_dir, **cfg)
        eKonf.env_set("WANDB_PROJECT", cfg.wandb_project)

        self._model = None
        self._dataset = None
        self._tokenizer = None
        self._diffuser = None

    @property
    def generate_config(self):
        return self.config.generate

    @generate_config.setter
    def generate_config(self, config):
        self.config.generate = config

    def generate_prompts(
        self,
        prompt=None,
        num_prompts_to_generate=5,
        batch_name=None,
        generate_images=True,
        num_samples=3,
        **kwargs,
    ):
        self.load_config()
        cfg = self.generate_config
        cfg.num_prompts_to_generate = num_prompts_to_generate
        cfg = eKonf.merge(cfg, kwargs)

        tokenizer = self.tokenizer
        model = self.model

        if prompt is None:
            prompt = cfg.prompt
        if prompt is None:
            prompt = self.start_token
        cfg.prompt = prompt

        encoded_prompt = tokenizer(
            prompt, add_special_tokens=False, return_tensors="pt"
        ).input_ids.to(model.device)

        output_sequences = model.generate(
            input_ids=encoded_prompt,
            max_length=cfg.max_prompt_length,
            min_length=cfg.min_prompt_length,
            temperature=cfg.temperature,
            top_k=cfg.top_k,
            top_p=cfg.top_p,
            do_sample=True,
            num_return_sequences=cfg.num_prompts_to_generate,
            pad_token_id=tokenizer.pad_token_id,  # gets rid of warning
        )

        tokenized_start_token = tokenizer.encode(self.start_token)

        generated_prompts = []
        for i, generated_sequence in enumerate(output_sequences):
            if not cfg.check_start_token:
                tokens = list(generated_sequence)
            else:
                tokens = []
                for i, s in enumerate(generated_sequence):
                    if s in tokenized_start_token and i != 0:
                        if len(tokens) >= cfg.min_prompt_length:
                            break
                    tokens.append(s)

            text = tokenizer.decode(
                tokens, clean_up_tokenization_spaces=True, skip_special_tokens=True
            )
            text = (
                text.strip().replace("\n", " ").replace("/", ",")
            )  # / remove slash. It causes problems in namings
            generated_prompts.append(text)
            if self.verbose:
                log.info(f"Prompt {i}: {text}")

        self.generated_prompts = generated_prompts
        self.generate_config = cfg
        self.save_config()

        if generate_images:
            self.generate_images(
                generated_prompts, batch_name=batch_name, num_samples=num_samples
            )
        return generated_prompts

    @property
    def generated_prompts(self):
        return self.config.generated_prompts

    @generated_prompts.setter
    def generated_prompts(self, value):
        self.config.generated_prompts = value

    def generate_images(
        self,
        prompts=None,
        batch_name=None,
        num_samples=3,
        max_display_image_width=800,
        **kwargs,
    ):
        if prompts is None:
            prompts = list(self.generated_prompts)
        if prompts is None:
            log.warning("No prompts provided")
            return None
        if not isinstance(prompts, list):
            prompts = [prompts]

        if batch_name is None:
            batch_name = f"{self.name}_batch"
        batch_run_params = {"text_prompts": prompts}
        batch_run_pairs = [["text_prompts"]]

        batch_results = self.diffuser.batch_imagine(
            batch_name=batch_name,
            batch_run_params=batch_run_params,
            batch_run_pairs=batch_run_pairs,
            num_samples=num_samples,
            max_display_image_width=max_display_image_width,
            **kwargs,
        )

        return batch_results

    @property
    def diffuser(self):
        if self._diffuser is None:
            self._diffuser = StableDiffusion()
        return self._diffuser

    def imagine(
        self,
        text_prompts=None,
        batch_name=None,
        num_samples=1,
        **imagine_args,
    ):
        if batch_name is None:
            batch_name = self.name
        return self.diffuser.imagine(
            text_prompts=text_prompts,
            batch_name=batch_name,
            num_samples=num_samples,
            **imagine_args,
        )

    @property
    def model(self):
        if self._model is None:
            self.load_model()
        return self._model

    @property
    def model_name(self):
        return self.config.model.model_name

    @model_name.setter
    def model_name(self, model_name):
        self.config.model.model_name = model_name

    def load_model(self, model_name: str = None, model_dir: str = None, **kwargs):
        if model_name is None:
            model_name = self.model_name
        if model_dir is None:
            model_dir = self.model_dir
        if model_name is None:
            log.warning("No model provided")
            return None

        if self._model is None or model_name != self.model_name:

            model_path = Path(model_dir) / model_name
            log.info(f"Loading model from {model_path}")
            self._model = AutoModelForCausalLM.from_pretrained(model_path, **kwargs)
            self._model.to(self.device)
            if self.verbose:
                log.info(f"Model: {self.model}")
            self.model_name = model_name

        else:
            log.info("Model already loaded")

    def tokenize(
        self,
        dataset,
        prompt_column,
        shuffle_prompts=False,
        num_shuffles=4,
        batch_size=10000,
        split="train",
        padding="max_length",
    ):
        tokenizer = self.tokenizer
        if tokenizer is None:
            log.warning("No tokenizer provided")
            return None
        if isinstance(dataset, DatasetDict):
            dataset = dataset[split]

        def _tokenize(dataset):
            # Remove empty lines
            dataset[prompt_column] = [
                self.start_token + prompt.strip() + self.end_token
                for prompt in dataset[prompt_column]
                if len(prompt.strip()) > 0
            ]

            outputs = tokenizer(
                dataset[prompt_column],
                padding=padding,
                truncation=True,
                # return_special_tokens_mask=True,
                max_length=tokenizer.model_max_length,
            )
            return {"input_ids": outputs["input_ids"]}

        def _tokenize_shuffled(dataset):
            block_size = tokenizer.model_max_length
            prompts = dataset[prompt_column]

            shuffled_prompts = ""
            for _ in range(num_shuffles):
                random.shuffle(prompts)
                concat = "".join(
                    [
                        self.start_token + prompt.strip() + " " + self.end_token
                        for prompt in prompts
                        if len(prompt.strip()) > 0
                    ]
                )
                shuffled_prompts += concat

            input_ids = tokenizer(shuffled_prompts)["input_ids"]

            result = []
            for i in range(
                0, len(input_ids) - block_size + 1, block_size
            ):  # Truncate in block of block_size
                result.append(
                    tokenizer.build_inputs_with_special_tokens(
                        input_ids[i : i + block_size]
                    )
                )
            return {"input_ids": result}

        if shuffle_prompts:
            _tokenize_function = _tokenize_shuffled
        else:
            _tokenize_function = _tokenize

        tokenized_dataset = dataset.map(
            _tokenize_function,
            batched=True,
            batch_size=batch_size,
            drop_last_batch=False,
            remove_columns=[prompt_column],
        )
        return tokenized_dataset

    @property
    def model_config(self):
        return self.config.model

    @model_config.setter
    def model_config(self, value):
        self.config.model = value

    @property
    def training_dir(self):
        return self.output_dir / "training"

    @property
    def train_config(self):
        return self.config.train

    @train_config.setter
    def train_config(self, value):
        self.config.train = value

    @property
    def trainer_config(self):
        return self.config.trainer

    @trainer_config.setter
    def trainer_config(self, value):
        self.config.trainer = value

    def train(
        self,
        model_name=None,
        pretrained_model_name_or_path=None,
        dataset=None,
        prompt_column=None,
        split=None,
    ):
        if model_name is None:
            model_name = self.model_name

        if model_name is None:
            log.warning("No model provided")
            return None

        if pretrained_model_name_or_path is None:
            pretrained_model_name_or_path = (
                self.train_config.pretrained_model_name_or_path
            )
        else:
            self.train_config.pretrained_model_name_or_path = (
                pretrained_model_name_or_path
            )
        if prompt_column is None:
            prompt_column = self.train_config.prompt_column
        else:
            self.train_config.prompt_column = prompt_column
        if split is None:
            split = self.train_config.split
        else:
            self.train_config.split = split

        # load up the dataset
        if dataset is None:
            dataset = self.dataset

        model_train_dir = self.training_dir / model_name
        cache_dir = self.cache_dir

        shuffle_prompts = self.train_config.shuffle_prompts
        num_shuffles = self.train_config.num_shuffles
        batch_size = self.train_config.batch_size
        padding = self.train_config.padding

        tokenizer = self.tokenizer
        tokenized_dataset = self.tokenize(
            dataset=dataset,
            prompt_column=prompt_column,
            shuffle_prompts=shuffle_prompts,
            num_shuffles=num_shuffles,
            batch_size=batch_size,
            split=split,
            padding=padding,
        )

        # train the model
        self.trainer_config.seed = self.seed
        self.trainer_config.run_name = model_name
        training_args = TrainingArguments(
            output_dir=model_train_dir,
            **self.trainer_config,
        )

        model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path, cache_dir=cache_dir
        )
        model.resize_token_embeddings(len(tokenizer))
        data_collator = DataCollatorForLanguageModeling(tokenizer=tokenizer, mlm=False)
        trainer = Trainer(
            model=model,
            tokenizer=tokenizer,
            args=training_args,
            data_collator=data_collator,
            train_dataset=tokenized_dataset,
        )
        trainer.train()

        wandb.finish()
        trainer = None

        self.generated_prompts = None
        self.save_model(model, model_name=model_name)

    def save_model(self, model, model_name=None, model_dir=None, **kwargs):
        if model is None:
            log.warning("No model provided")
            return None

        if model_name is None:
            model_name = self.model_name
        if model_dir is None:
            model_dir = self.model_dir

        if model_name is None:
            log.warning("No model provided")
            return None

        model_path = Path(model_dir) / model_name
        model_path.mkdir(parents=True, exist_ok=True)
        log.info(f"Saving model to {model_path}")
        model.save_pretrained(model_path, **kwargs)
        self._model = model
        self._model.to(self.config.device)
        self.model_name = model_name
        self.save_config()

    @property
    def dataset_config(self):
        return self.config.dataset

    @dataset_config.setter
    def dataset_config(self, value):
        self.config.dataset = value

    @property
    def dataset(self):
        if self._dataset is None:
            self.load_dataset()
        return self._dataset

    def load_dataset(
        self,
        path=None,
        download_mode=None,
        **kwargs,
    ):
        if path is None:
            path = self.dataset_config.path
        else:
            self.dataset_config.path = path
        if download_mode is None:
            download_mode = self.dataset_config.download_mode
        else:
            self.dataset_config.download_mode = download_mode

        self._dataset = datasets.load_dataset(
            path,
            cache_dir=self.cache_dir,
            download_mode=download_mode,
            **kwargs,
        )
        if self.verbose:
            log.info(f"Dataset: {self._dataset}")
        return self._dataset

    @property
    def tokenize_config(self):
        return self.config.tokenize

    @property
    def special_tokens(self):
        return {
            "bos_token": self.start_token,
            "eos_token": self.end_token,
            "pad_token": self.pad_token,
        }

    @property
    def tokenizer(self):
        if self._tokenizer is None:
            self.load_tokenizer()
        return self._tokenizer

    def load_tokenizer(self, pretrained_model_name_or_path=None, **kwargs):
        if pretrained_model_name_or_path is None:
            pretrained_model_name_or_path = (
                self.train_config.pretrained_model_name_or_path
            )

        self._tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path,
            cache_dir=self.cache_dir,
            **kwargs,
        )
        self._tokenizer.add_special_tokens(self.special_tokens)
        if self.verbose:
            log.info(f"Tokenizer: {self._tokenizer}")



import logging
import numpy as np
import pandas as pd
import random
from enum import Enum
from pathlib import Path
from typing import Optional, Union, List
from pydantic import BaseModel, validator, PrivateAttr
from ekorpkit.batch import BaseBatchConfig
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class ImagineMode(str, Enum):
    """Diffuse mode"""

    GENERATE = "generate"
    INPAINT = "inpaint"
    STITCH = "stitch"


class SchedulerType(str, Enum):
    K_LMS = "k_lms"
    PNDM = "pndm"
    DDIM = "ddim"


class ImagineConfig(BaseModel):
    text_prompts: dict = None
    num_inference_steps: int = 100
    num_samples: int = 2
    num_images_per_prompt: int = 2
    init_image: str = None
    init_strength: Optional[float] = None
    init_max_pixels: int = 262144
    guidance_scale: float = 7.5
    width: int = 512
    height: int = 512
    seed: Optional[int] = None
    increase_seed_by: Optional[int] = 1
    clear_output: bool = True
    display_image: bool = True
    save_image: bool = True
    save_image_config: bool = True
    display_collage: bool = False
    save_collage: bool = True
    image_ext: str = "png"
    _text_series: Optional[pd.Series] = PrivateAttr(None)

    def __init__(self, **values):
        text_prompts = values.get("text_prompts")
        if isinstance(text_prompts, str):
            text_prompts = {0: [text_prompts]}
        elif isinstance(text_prompts, list):
            text_prompts = {0: text_prompts}
        values["text_prompts"] = text_prompts
        super().__init__(**values)

    @validator("width")
    def width_divisible_by_8(cls, v):
        if v % 8 != 0:
            log.warning(f"Width {v} is not divisible by 8. Setting to {v // 8 * 8}")
        return (v // 8) * 8

    @validator("height")
    def height_divisible_by_8(cls, v):
        if v % 8 != 0:
            log.warning(f"Height {v} is not divisible by 8. Setting to {v // 8 * 8}")
        return (v // 8) * 8

    @validator("num_images_per_prompt")
    def num_images_per_prompt_less_than_num_samples_and_divisible(cls, v, values):
        num_samples = values["num_samples"]
        if v > num_samples:
            msg = (
                f"num_images_per_prompt {v} is greater than num_samples {num_samples}. "
            )
            msg += f"Setting to {num_samples}"
            log.warning(msg)
            v = num_samples
        if num_samples % v != 0:
            msg = f"num_samples {num_samples} is not divisible by num_images_per_prompt {v}. "
            msg += f"Setting to {num_samples // v * v}"
            log.warning(msg)
            v = num_samples // v * v
        return v

    @property
    def num_prompts(self):
        return len(self.text_prompts)

    @property
    def num_iterations(self):
        return int(self.num_samples / self.num_images_per_prompt)

    def get_prompt(self, sample_num: int = 0) -> List[str]:
        text_series = self.prompt_series
        if text_series is not None and sample_num >= len(text_series):
            text_prompt = text_series[-1]
        elif text_series is not None:
            text_prompt = text_series[sample_num]
        else:
            text_prompt = []
        return text_prompt

    @property
    def prompt_series(self):
        """Get the prompt series"""
        if self._text_series is not None:
            return self._text_series
        text_prompts = self.text_prompts
        text_series = (
            split_prompts(text_prompts, self.num_samples) if text_prompts else None
        )
        self._text_series = text_series
        return text_series

    def get_str_prompt(self):
        return get_str_prompt(self.text_prompts)


class StableImagineConfig(ImagineConfig):
    mode: ImagineMode = ImagineMode.GENERATE
    scheduler: SchedulerType = SchedulerType.K_LMS
    NSFW_retry: int = 0
    mask_image: str = None
    inpaint_strength: float = 7.5


class CollageConfig(BaseModel):
    clear_output = True
    show_prompt = True
    ncols = 3
    max_images = 12
    collage_width = 1200
    padding: int = 5
    bg_color: str = "black"
    crop_to_min_size = False
    show_filename = False
    filename_offset = (5, 5)
    fontname: str = None
    fontsize: int = 12
    fontcolor: str = "#000"


class AnimMode(str, Enum):
    """Animation mode"""

    NONE = "None"
    ANIM_2D = "2D"
    ANIM_3D = "3D"
    VIDEO_INPUT = "Video Input"


class BatchConfig(BaseBatchConfig):
    max_display_image_width: int = 1000
    collage_dirname = "collages"
    collage_filesfx = "collage.png"

    def get_imagine_config(self, pair_args: dict = None, **imagine_args):
        args = imagine_args.copy()
        if pair_args is not None:
            args["num_samples"] = self.num_samples
            args.update(pair_args)
        return args

    @property
    def collage_dir(self):
        collage_dir = self.batch_dir / self.collage_dirname
        collage_dir.mkdir(parents=True, exist_ok=True)
        return collage_dir

    @property
    def collage_filename(self):
        return f"{self.file_prefix}_{self.collage_filesfx}"

    @property
    def collage_filepath(self):
        return self.collage_dir / self.collage_filename


class BatchImagineResult(BaseModel):
    run_configs: dict
    collage_filepaths: dict
    batch_prompts: dict


class ImagineResult(BaseModel):
    batch_num: int = None
    batch_args: dict = None
    config_file: str = None
    image_filepaths: list = []
    stitched_image_path: str = None


class BatchRunCfg(BaseModel):
    batch_name: str
    batch_run_pair: dict
    imagine_results: list = []
    run_config_path: Path = None

    def __init__(self, **values):
        run_config_path = values.get("run_config_path")
        if run_config_path is not None:
            run_config_path = Path(run_config_path)
            if run_config_path.exists():
                run_config = eKonf.load(run_config_path)
                values.update(eKonf.to_dict(run_config))
        values["run_config_path"] = run_config_path
        super().__init__(**values)

    @validator("imagine_results")
    def validate_imagine_results(cls, v):
        return [ImagineResult(**d) if isinstance(d, dict) else d for d in v]

    def append(self, batch_args, imagine_result: ImagineResult):
        imagine_result.batch_args = batch_args
        self.imagine_results.append(imagine_result)

    @property
    def collage_dir(self):
        return self.run_config_path.parent.parent

    def collage_filename(self, suffix=None):
        sfx = f"_{suffix}" if suffix else ""
        return self.run_config_path.name.replace(".yaml", f"{sfx}.png")

    def collage_filepath(self, suffix=None):
        return self.collage_dir / self.collage_filename(suffix)


class BatchRunConfig(BatchConfig):
    batch_run_params: dict = None
    batch_run_pairs: list = None
    num_samples: int = 1
    num_images_per_prompt: int = 1
    clear_output: bool = True
    display_collage: bool = False
    save_collage: bool = True
    run_config_file = "run_configs.yaml"

    @validator("batch_run_params")
    def batch_run_params_is_dict(cls, v):
        if v is None:
            return {}
        if not isinstance(v, dict):
            raise ValueError("batch_run_params must be a dict")
        if len(v) < 1:
            raise ValueError("batch_run_params must have at least 1 element")
        for _, val in v.items():
            if not isinstance(v, (list, tuple)):
                val = [val]
        return v

    @validator("batch_run_pairs")
    def batch_run_pairs_is_list_of_tuples(cls, v, values):
        if v is None:
            log.info("No batch_run_pairs provided. Using each key in batch_run_params")
            batch_run_params = values.get("batch_run_params")
            if batch_run_params:
                batch_run_params = cls.batch_run_params_is_dict(batch_run_params)
                v = [[arg_name] for arg_name in batch_run_params.keys()]
            else:
                v = [[]]
        if not isinstance(v, list):
            raise ValueError("batch_run_pairs must be a list")
        # Check that each element is a list or tuple, if an element is a string, convert to list
        for i, pair in enumerate(v):
            if isinstance(pair, str):
                v[i] = [pair]
            elif not isinstance(pair, (list, tuple)):
                raise ValueError("batch_run_pairs must be a list of lists or tuples")
        return v

    @property
    def batch_run_configs(self):
        for run_pair in self.batch_run_pairs:
            batch_run_pair = {name: self.batch_run_params[name] for name in run_pair}
            batch_run_name = self.batch_name + "_" + "_".join(run_pair)
            yield BatchRunCfg(
                batch_name=batch_run_name,
                batch_run_pair=batch_run_pair,
                results=[],
            )

    def get_run_args(self, batch_run_pair):
        return eKonf.dict_product(batch_run_pair)

    def get_imagine_config(self, arg_pair: dict = None, **imagine_args):
        args = imagine_args.copy()
        if arg_pair is not None:
            # args["num_samples"] = self.num_samples
            # args["increase_seed_by"] = 0
            if "text_prompts" not in arg_pair:
                args["display_collage"] = False
            args.update(arg_pair)
        return args

    def save_run_config(self, run_config: BatchRunCfg):
        """Save the settings"""
        batch_name = run_config.batch_name
        filename = f"{batch_name}({self.batch_num})_{self.run_config_file}"
        config_path = self.config_dir / filename
        log.info(f"Saving batch run config to {config_path}")
        eKonf.save(run_config.dict(), config_path)
        return config_path

    def load_run_config(self, run_config_path):
        """Load the settings"""
        log.info(f"Loading batch run config from {run_config_path}")
        run_config = eKonf.load(run_config_path)
        return BatchRunCfg(**run_config)


class RunConfig(BaseModel):
    batch: BatchConfig
    imagine: ImagineConfig
    collage: CollageConfig = CollageConfig()

    def __init__(self, **values):
        super().__init__(**values)
        self.set_seed()

    @property
    def batch_file_prefix(self):
        return self.batch.file_prefix

    @property
    def batch_dir(self):
        return self.batch.batch_dir

    @property
    def image_ext(self):
        return self.imagine.image_ext

    def set_seed(self, seed: int = None):
        if seed is None:
            seed = self.batch.seed
        if seed is None or seed < 0:
            random.seed()
            seed = random.randint(0, 2**32 - 1)
        if self.imagine.seed is None:
            log.info(f"Setting imagine seed to {seed}")
            self.imagine.seed = seed
        else:
            log.info(f"Seed already set to {self.imagine.seed}")

    @property
    def image_filepattern(self):
        return f"{self.batch_file_prefix}*{self.image_ext}"

    @property
    def image_filepaths(self):
        return sorted(self.batch_dir.glob(self.image_filepattern))

    def get_sample_path(
        self, name: Union[int, str], seed: int = None, ext: str = None
    ) -> str:
        if ext is None:
            ext = self.image_ext
        if isinstance(name, int):
            name = f"{name:04d}"
        else:
            name = str(name)
        _seed = "" if seed is None else f"_{seed}"
        filename = f"{self.batch.file_prefix}_{name}{_seed}.{ext}"
        return str(self.batch.batch_dir / filename)

    def save(self, img, name: Union[int, str], seed: int = None):
        img_filepath = self.get_sample_path(name, seed, ext=self.image_ext)
        if self.imagine.save_image:
            img.save(img_filepath)
            log.info(f"Saved image to {img_filepath}")
        if self.imagine.save_image_config:
            self.save_config(name, seed)
        return img_filepath

    def save_config(self, name: Union[int, str], seed: int = None):
        rc_filepath = self.get_sample_path(name, seed, ext="json")
        with open(rc_filepath, "w") as f:
            f.write(self.json(indent=4, sort_keys=True, ensure_ascii=False))
        log.info(f"Saved run config to {rc_filepath}")


class StableRunConfig(RunConfig):
    imagine: StableImagineConfig


def split_prompts(prompts, max_frames):
    prompt_series = pd.Series([np.nan for a in range(max_frames)])
    for i, prompt in prompts.items():
        if isinstance(prompt, str):
            prompt = [prompt]
        prompt_series[i] = prompt
    # prompt_series = prompt_series.astype(str)
    prompt_series = prompt_series.ffill().bfill()
    return prompt_series


def get_str_prompt(prompts):
    prompts = eKonf.to_dict(prompts)
    if isinstance(prompts, str):
        return prompts
    elif isinstance(prompts, list):
        return ", ".join(prompts)
    elif isinstance(prompts, dict):
        if 0 in prompts:
            prompts = prompts[0]
        else:
            prompts = prompts.values()[0]
        return get_str_prompt(prompts)



import os
import json
import logging
import numpy as np
from scipy.special import digamma
from typing import Iterator, Optional, Union
from .base import Model
from ..trainers.unigram import UnigramTrainer
from ..utils.trie import Trie


log = logging.getLogger(__name__)


class Unigram(Model):
    """
    An implementation of the Unigram algorithm
    Args:
        vocab (:obj:`List[Tuple[str, float]]`, `optional`):
            A list of vocabulary items and their relative score [("am", -0.2442),...]
    """

    def __init__(self, vocab=None):
        super().__init__(vocab)

    def initialize_vocab(self, vocab, **kwargs):
        self.vocab = {}
        self.token2id = {}
        self.id2token = {}
        self.trie = None
        self.max_piece_length = None
        if vocab:
            self.vocab = vocab
            self.token2id = {token: i for i, token in enumerate(vocab)}
            self.id2token = {v: k for k, v in self.token2id.items()}
            self.trie, self.max_piece_length = self.initialize_trie(vocab)

    def initialize_trie(self, tokens):
        trie = Trie()
        norm = sum(list(tokens.values()))
        logsum = digamma(norm)

        maxlen = 0
        for tok, val in tokens.items():
            trie.add(tok, digamma(val) - logsum)
            maxlen = max(maxlen, len(tok))

        return trie, maxlen

    def generalized_forward_step(self, text, trie, nbest_size=1):
        N = len(text)
        d = [-np.inf] * (N + 1)
        p = [None] * (N + 1)
        d[0] = 0
        for i in range(1, N + 1):
            d_queue = []
            p_queue = []
            for j in range(max(i - self.max_piece_length, 0), i):
                final_token = text[j:i]
                final_value = trie.get_value(final_token)
                if final_value:
                    curr_d = d[j] + final_value
                    curr_p = len(final_token)
                    d[i] = max(d[i], curr_d)
                    d_queue.append(curr_d)
                    p_queue.append(curr_p)
            ids = np.argsort(d_queue)[-nbest_size:]
            p[i] = [p_queue[z] for z in ids]
        return p

    def generalized_backward_step(self, text, p):
        idx = len(p)
        tokenization = []
        while idx > 1:
            back_steps = np.random.choice(p[idx - 1])
            next_idx = idx - back_steps
            tok = text[next_idx - 1 : idx - 1]
            tokenization.append(tok)
            idx = next_idx
        tokenization = list(reversed(tokenization))
        return tokenization

    def tokenize(self, sequence, nbest_size=1):
        """
        Tokenize a sequence
        Args:
            sequence (:obj:`str`):
                A sequence to tokenize
            nbest_size (:obj:`int`, `optional`, defaults to 1):
        Returns:
            A :obj:`List` of :class:`~tokenizers.Token`: The generated tokens
        """
        sequence = self.pre_tokenize(sequence)
        text = "".join(sequence)
        if self.trie is None:
            raise ValueError("Trainer has not yet been fit. Cannot tokenize.")
        p = self.generalized_forward_step(text, self.trie, nbest_size)
        tokenization = self.generalized_backward_step(text, p)
        return tokenization

    def train_from_iterator(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        trainer: UnigramTrainer = None,
        length: Optional[int] = None,
    ):
        """Train the model using the given iterator"""

        trainer.normalizer = self.normalizer
        trainer.pre_tokenizer = self.pre_tokenizer
        vocab = trainer.train(iterator, length=length)
        self.initialize_vocab(vocab)

    @classmethod
    def from_file(cls, vocab, **kwargs):
        """
        Instantiate a Unigram model from the given files.

        This method is roughly equivalent to doing::

           vocab, merges = Unigram.read_file(vocab_filename, merges_filename)
           bpe = Unigram(vocab, merges)

        Args:
            vocab (:obj:`str`):
                The path to a :obj:`vocab.json` file

        Returns:
            :class:`~tokenizers.models.Unigram`: An instance of Unigram loaded from these files
        """
        vocab = cls.read_file(vocab)
        return cls(vocab, **kwargs)

    @staticmethod
    def read_file(vocab):
        """
        Read a :obj:`vocab.json` file

        This method provides a way to read and parse the content of these files,
        returning the relevant data structures. If you want to instantiate some BPE models
        from memory, this method gives you the expected input from the standard files.

        Args:
            vocab (:obj:`str`):
                The path to a :obj:`vocab.json` file

        Returns:
            A :obj:`Tuple` with the vocab and the merges:
                The vocabulary and merges loaded into memory
        """
        with open(vocab, "r") as f:
            vocab = json.load(f)
        return vocab

    def save(self, folder, prefix=None, pretty: bool = False):
        """
        Save the current model

        Save the current model in the given folder, using the given prefix for the various
        files that will get created.
        Any file with the same name that already exists in this folder will be overwritten.

        Args:
            folder (:obj:`str`):
                The path to the target folder in which to save the various files

            prefix (:obj:`str`, `optional`):
                An optional prefix, used to prefix each file name

        Returns:
            :obj:`List[str]`: The list of saved files
        """
        if prefix is not None:
            folder = os.path.join(folder, prefix)
        vocab_filename = os.path.join(folder, "vocab.json")
        if not os.path.exists(folder):
            os.makedirs(folder)
        indent = 2 if pretty else None
        json.dump(self.vocab, open(vocab_filename, "w"), indent=indent)
        return [vocab_filename]



import logging
import torch
from diffusers import StableDiffusionPipeline, StableDiffusionInpaintPipeline
from diffusers.schedulers.scheduling_ddim import DDIMScheduler
from diffusers.schedulers.scheduling_lms_discrete import LMSDiscreteScheduler
from diffusers.schedulers.scheduling_pndm import PNDMScheduler
from tqdm.auto import tqdm
from PIL import Image, ImageDraw
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from .base import BaseModel
from .config import (
    ImagineMode,
    BatchConfig,
    StableImagineConfig,
    StableRunConfig,
    SchedulerType,
    CollageConfig,
    ImagineResult,
)


log = logging.getLogger(__name__)


class StableDiffusion(BaseModel):
    def __init__(
        self, hf_user_access_token=None, root_dir=None, config_name="default", **args
    ):
        cfg = eKonf.compose(f"model/stable_diffusion={config_name}")
        if hf_user_access_token is not None:
            args["hf_user_access_token"] = hf_user_access_token
        cfg = eKonf.merge(cfg, args)
        super().__init__(root_dir=root_dir, **cfg)

        self.pipes = {}
        self.inpaint_pipe = None
        self.generator = None

        if not self.hf_user_access_token:
            self.login_hf_hub()
        if self.autoload:
            self.load_diffusers()

    def login_hf_hub(self):
        from huggingface_hub import notebook_login

        if eKonf.is_notebook():
            notebook_login()
        else:
            raise ValueError(
                "huggingface_hub.notebook_login() is only available in notebook, set hf_user_access_token manually"
            )

    @property
    def hf_user_access_token(self):
        return self.config.hf_user_access_token or eKonf.osenv("HF_USER_ACCESS_TOKEN")

    def imagine(
        self,
        text_prompts=None,
        batch_name=None,
        batch_num=None,
        mode: ImagineMode = None,
        num_samples=None,
        return_including_init_image=False,
        **imagine_args,
    ):
        """Generate images"""

        if text_prompts is not None:
            imagine_args.update(dict(text_prompts=text_prompts))
        if mode is not None:
            imagine_args.update(dict(mode=mode))
        if num_samples is not None:
            imagine_args.update(dict(num_samples=num_samples))
        log.info("> loading config...")
        config = self.load_config(
            batch_name=batch_name,
            batch_num=batch_num,
            imagine=imagine_args,
        )
        config_file = self.save_config(config)
        rc = self.get_run_config(config)
        # run_cfg.set_seed()

        imagine_rst = ImagineResult(batch_num=rc.batch.batch_num)
        sample_imagepaths = []
        with elapsed_timer(format_time=True) as elapsed:
            if rc.imagine.mode == ImagineMode.GENERATE:
                sample_imagepaths = self.generate_images(rc)
            elif rc.imagine.mode == ImagineMode.INPAINT:
                sample_imagepaths = self.inpaint_images(rc, return_including_init_image)
            elif rc.imagine.mode == ImagineMode.STITCH:
                rc.imagine.num_images_per_prompt = 1
                sample_imagepaths = self.generate_images(rc)
                stitched_image_path = self.stitch_images(sample_imagepaths, rc)
                imagine_rst.stitched_image_path = stitched_image_path
                rc.imagine.save_collage = False
                rc.imagine.display_collage = False

            log.info(" >> elapsed time to imagine: {}".format(elapsed()))

        if len(sample_imagepaths) == 0:
            log.info(" >> no images generated")
            return imagine_rst

        if rc.imagine.save_collage or rc.imagine.display_collage:
            if rc.imagine.clear_output:
                eKonf.clear_output(wait=True)
            self.collage(
                images_or_uris=sample_imagepaths,
                save_collage=rc.imagine.save_collage,
                display_collage=rc.imagine.display_collage,
                **rc.collage.dict(),
            )

        imagine_rst.image_filepaths = sample_imagepaths
        imagine_rst.config_file = config_file
        return imagine_rst

    def get_run_config(self, config):
        batch = BatchConfig(output_dir=config.path.output_dir, **config.batch)
        imagine = StableImagineConfig(**config.imagine)
        collage = CollageConfig(**config.collage)
        rc = StableRunConfig(batch=batch, imagine=imagine, collage=collage)
        return rc

    def generate_images(self, rc: StableRunConfig):
        # Generate images
        cfg = rc.imagine
        images = []
        image_num = 0
        with torch.autocast("cuda"):
            for i in tqdm(range(cfg.num_iterations)):
                log.info(f"> generating image {image_num+1}/{cfg.num_samples}")
                imgs = self.generating(
                    prompt=cfg.get_prompt(i),
                    width=cfg.width,
                    height=cfg.height,
                    guidance_scale=cfg.guidance_scale,
                    num_images_per_prompt=cfg.num_images_per_prompt,
                    num_inference_steps=cfg.num_inference_steps,
                    seed=cfg.seed,
                )
                for img in imgs:
                    img_path = rc.save(img, image_num, seed=cfg.seed)
                    if cfg.display_image:
                        if cfg.clear_output:
                            eKonf.clear_output(wait=True)
                        if rc.batch.max_display_image_width is not None:
                            img = eKonf.scale_image(
                                img, max_width=rc.batch.max_display_image_width
                            )
                        eKonf.display(img)
                    images.append(img_path)
                    image_num += 1
                if cfg.increase_seed_by:
                    cfg.seed += cfg.increase_seed_by
        return images

    def generating(
        self,
        prompt,
        width=512,
        height=512,
        guidance_scale=7.5,
        num_images_per_prompt=1,
        num_inference_steps=50,
        **kwargs,
    ):
        """Generate images from a prompt"""
        pipe = self.get_pipe("generate")
        images = pipe(
            prompt=prompt,
            width=width,
            height=height,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            generator=self.generator,
            **kwargs,
        ).images
        return images

    def inpaint_images(
        self,
        rc: StableRunConfig,
        return_including_init_image=False,
    ):
        # Inpaint images
        cfg = rc.imagine
        images = []
        image_num = 0

        init_image = eKonf.load_image(cfg.init_image)
        mask_image = eKonf.load_image(cfg.mask_image)

        with torch.autocast("cuda"):
            for i in tqdm(range(cfg.num_iterations)):
                log.info(f"> generating image {image_num+1}/{cfg.num_samples}")
                imgs = self.inpainting(
                    prompt=cfg.get_prompt(i),
                    init_image=init_image,
                    mask_image=mask_image,
                    guidance_scale=cfg.inpaint_strength,
                    num_images_per_prompt=cfg.num_images_per_prompt,
                    num_inference_steps=cfg.num_inference_steps,
                    seed=cfg.seed,
                )
                for img in imgs:
                    img_path = rc.save(img, image_num, seed=cfg.seed)
                    if cfg.display_image:
                        if cfg.clear_output:
                            eKonf.clear_output(wait=True)
                        if rc.batch.max_display_image_width is not None:
                            img = eKonf.scale_image(
                                img, max_width=rc.batch.max_display_image_width
                            )
                        eKonf.display(img)
                    images.append(img_path)
                    image_num += 1
                if cfg.increase_seed_by:
                    cfg.seed += cfg.increase_seed_by
        # insert initial image in the list so we can compare side by side
        if return_including_init_image:
            images.insert(0, cfg.init_image)
        return images

    def inpainting(
        self,
        prompt,
        init_image,
        mask_image,
        guidance_scale=7.5,
        num_images_per_prompt=1,
        num_inference_steps=50,
        **kwargs,
    ):
        """Inpainting"""
        width, height = init_image.size
        if height % 8 != 0 or width % 8 != 0:
            height = (height // 8) * 8
            width = (width // 8) * 8
            init_image = init_image.crop((0, 0, width, height))
            mask_image = mask_image.crop((0, 0, width, height))

        pipe = self.get_pipe("inpaint")
        images = pipe(
            prompt=prompt,
            image=init_image,
            mask_image=mask_image,
            width=width,
            height=height,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            generator=self.generator,
            **kwargs,
        ).images
        return images

    def stitch_images(
        self,
        image_filepaths,
        rc: StableRunConfig,
        **kwargs,
    ):
        # Stitch images
        cfg = rc.imagine
        images = [
            eKonf.load_image(image_filepath) for image_filepath in image_filepaths
        ]

        with torch.autocast("cuda"):
            output = self.stitching(
                images,
                prompt=cfg.get_prompt(0),
                width=cfg.width,
                height=cfg.height,
                inpaint_strength=cfg.inpaint_strength,
                num_images_per_prompt=cfg.num_images_per_prompt,
                num_inference_steps=cfg.num_inference_steps,
                seed=cfg.seed,
                max_display_image_width=rc.batch.max_display_image_width,
            )

        # add borders
        w, h = output.size
        panorama = Image.new("RGB", (w, 3 * h))
        panorama.paste(output, (0, h))
        img_path = rc.save(panorama, "panorama")
        if cfg.display_image:
            if cfg.clear_output:
                eKonf.clear_output(wait=True)
            if rc.batch.max_display_image_width is not None:
                panorama = eKonf.scale_image(
                    panorama, max_width=rc.batch.max_display_image_width
                )
            eKonf.display(panorama)
        return img_path

    def stitching(
        self,
        images,
        prompt,
        width=512,
        height=512,
        inpaint_strength=7.5,
        num_images_per_prompt=1,
        num_inference_steps=50,
        max_display_image_width=None,
        **kwargs,
    ):
        """Stitching images"""

        output = images[0]
        pipe = self.get_pipe("inpaint")

        for img in tqdm(images[1:] + [images[0]]):
            w, h = output.size
            new = Image.new("RGB", (width, height))
            new.paste(output, (-w + width // 2, 0))
            new.paste(img, (width // 2, 0))

            msk = Image.new("L", (width, height))
            drw = ImageDraw.Draw(msk)
            drw.rectangle((width // 4, 0, 3 * width // 4, height), fill=255)
            _width, _height = new.size
            merged = pipe(
                prompt=prompt,
                image=new,
                mask_image=msk,
                width=_width,
                height=_height,
                guidance_scale=inpaint_strength,
                num_images_per_prompt=num_images_per_prompt,
                num_inference_steps=num_inference_steps,
                generator=self.generator,
                **kwargs,
            ).images[0]
            eKonf.clear_output(wait=True)
            _img = eKonf.scale_image(merged, max_width=max_display_image_width)
            eKonf.display(_img)

            new = Image.new("RGB", (w + width, height))
            new.paste(output, (0, 0))
            new.paste(merged, (w - width // 2, 0))
            img = img.crop((width // 2, 0, width, height))
            new.paste(img, (w + width // 2, 0))
            output = new

        w, h = output.size
        output = output.crop((width // 2, 0, w - width // 2, h))
        return output

    def get_pipe(self, name: str):
        pipe = self.pipes.get(name)
        if pipe is None:
            self.load_diffusers(name)
            pipe = self.pipes.get(name)
        if pipe is None:
            raise ValueError(f"pipe {name} not found")
        return pipe

    def load_diffusers(self, models=None):
        torch.set_grad_enabled(False)

        if models is None:
            models = self.model_config.keys()
        if isinstance(models, str):
            models = [models]
        for model in models:
            cfg = self.model_config[model]
            if cfg.pipeline == "StableDiffusionPipeline":
                DiffusionPipeline = StableDiffusionPipeline
            elif cfg.pipeline == "StableDiffusionInpaintPipeline":
                DiffusionPipeline = StableDiffusionInpaintPipeline
            else:
                raise ValueError(f"pipeline {cfg.pipeline} not found")

            self.pipes[model] = DiffusionPipeline.from_pretrained(
                pretrained_model_name_or_path=cfg.name,
                use_auth_token=self.hf_user_access_token,
                revision=cfg.revision,
                torch_dtype=torch.float16,
                cache_dir=self.path.cache_dir,
            ).to(cfg.device)

    def load_config(self, batch_name=None, batch_num=None, **kwargs):
        """Load the settings"""
        config = super().load_config(
            batch_name=batch_name, batch_num=batch_num, **kwargs
        )

        self.generator = torch.Generator(device=self.device).manual_seed(self.seed)
        return config

    def compare_images(self, images, rows=1, cols=None, resize_ratio=1.0):
        import PIL

        if cols is None:
            cols = len(images) // rows

        if resize_ratio != 1.0:
            images = [
                img.resize(
                    (int(img.width * resize_ratio), int(img.height * resize_ratio))
                )
                for img in images
            ]
        w, h = images[0].size
        grid = PIL.Image.new("RGB", size=(cols * w, rows * h))
        # grid_w, grid_h = grid.size

        for i, img in enumerate(images):
            grid.paste(img, box=(i % cols * w, i // cols * h))
        return grid


def set_scheduler(pipe, scheduler: SchedulerType):
    if scheduler == SchedulerType.DDIM:
        ddim = DDIMScheduler(
            beta_start=0.00085,
            beta_end=0.012,
            beta_schedule="scaled_linear",
            clip_sample=False,
            set_alpha_to_one=False,
        )
        pipe.scheduler = ddim
    elif scheduler == SchedulerType.K_LMS:
        lms = LMSDiscreteScheduler(
            beta_start=0.00085, beta_end=0.012, beta_schedule="scaled_linear"
        )
        pipe.scheduler = lms
    elif scheduler == SchedulerType.PNDM:
        pndm = PNDMScheduler(
            beta_start=0.00085, beta_end=0.012, beta_schedule="scaled_linear"
        )
        pipe.scheduler = pndm
    else:
        raise ValueError(f"Unknown scheduler {scheduler}")



import gzip
import shutil
import os
import tarfile
import zipfile
import pycurl
from tqdm.auto import tqdm
from urllib import request
from ekorpkit import eKonf


def download_from_gcs(**cfg):
    """Download the data from source url, unless it's already here.

    Args:
        filename: string, name of the file in the directory.
        work_directory: string, path to working directory.
        source_url: url to download from if file doesn't exist.

    Returns:
        Path to resulting file.
    """
    import tensorflow as tf

    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)
    if not os.listdir(args.output_dir) or args.force_download:
        for filename, source_path in args.data_sources.items():
            output_path = f"{args.output_dir}/{filename}"
            print(f"Downloading {source_path} to {output_path}")
            tf.io.gfile.copy(source_path, output_path, overwrite=args.force_download)
    else:
        print(f" >> {args.output_dir} is not empty, skip downloading")


def download_from_web(**cfg):
    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)
    if not os.listdir(args.output_dir) or args.force_download:
        for filename, url in args.urls.items():
            filepath = f"{args.output_dir}/{filename}"
            print(f"Downloading {url} to {filepath}")
            # download_file(url, file_path)
            web_download(url, filepath, args.name, force_download=False)
        if args.extract_command:
            print(args.extract_command)
    else:
        print(f" >> {args.output_dir} is not empty, skip downloading")


def download_from_gdrive(**cfg):
    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)
    for filename, url in args.urls.items():
        file_path = f"{args.output_dir}/{filename}"
        gdrive_download(url, file_path, args.name, args.force_download)


def download_from_gdrive_untar(**cfg):
    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)
    for filename, url in args.urls.items():
        file_path = f"{args.output_dir}/{filename}"
        gdrive_download_untar(url, file_path, args.name, args.force_download)


def web_download(
    url, local_path, filename="", force_download=False, verbose=True, **kwargs
):
    if filename == "":
        filename = os.path.basename(local_path)
    site = request.urlopen(url)
    meta = site.info()
    remote_size = int(meta["Content-Length"])
    if (
        (not force_download)
        and os.path.exists(local_path)
        and (os.stat(local_path).st_size == remote_size)
    ):
        if verbose:
            print(f"[{filename}] is already downloaded at {local_path}")
        return None
    filename = os.path.basename(local_path)
    if not os.path.exists(os.path.dirname(local_path)):
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
    with tqdm(
        unit="B", unit_scale=True, miniters=1, desc=f"[{filename}] download {filename}"
    ) as t:

        def _reporthook(pbar):
            last_b = [0]

            def inner(b=1, bsize=1, tsize=None):
                """
                Args:
                    b (int, optional): Number of blocks just transferred [default: 1].
                    bsize (int, optional): Size of each block (in tqdm units) [default: 1].
                    tsize (int, optional): Total size (in tqdm units). If [default: None] remains unchanged.
                """
                if tsize is not None:
                    pbar.total = tsize
                pbar.update((b - last_b[0]) * bsize)
                last_b[0] = b

            return inner

        request.urlretrieve(url, filename=local_path, reporthook=_reporthook(t))


def curl(url, local_path, filename="", force_download=False, verbose=True, **kwargs):
    if filename == "":
        filename = os.path.basename(local_path)
    site = request.urlopen(url)
    meta = site.info()
    remote_size = int(meta["Content-Length"])
    if (
        (not force_download)
        and os.path.exists(local_path)
        and (os.stat(local_path).st_size == remote_size)
    ):
        if verbose:
            print(f"[{filename}] is already downloaded at {local_path}")
        return None
    filename = os.path.basename(local_path)
    if not os.path.exists(os.path.dirname(local_path)):
        os.makedirs(os.path.dirname(local_path), exist_ok=True)

    # create a progress bar and update it manually
    with tqdm(
        total=remote_size,
        unit="iB",
        unit_scale=True,
        desc=f"[{filename}] download {filename}",
    ) as pbar:
        # store dotal dl's in an array (arrays work by reference)
        total_dl_d = [0]

        def _status(download_t, download_d, upload_t, upload_d, total=total_dl_d):
            # increment the progress bar
            pbar.update(download_d - total[0])
            # update the total dl'd amount
            total[0] = download_d

        # download file using pycurl
        with open(local_path, "wb") as f:
            c = pycurl.Curl()
            c.setopt(c.URL, url)
            c.setopt(c.WRITEDATA, f)
            # follow redirects:
            c.setopt(c.FOLLOWLOCATION, True)
            # custom progress bar
            c.setopt(c.NOPROGRESS, False)
            c.setopt(c.XFERINFOFUNCTION, _status)
            c.perform()
            c.close()


def web_download_unzip(
    url, zip_path, filename="", force_download=False, verbose=True, **kwargs
):
    web_download(url, zip_path, filename, force_download)
    # assume that path/to/abc.zip consists path/to/abc
    data_path = zip_path[:-4]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    data_root = os.path.dirname(zip_path)
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(data_root)
    if verbose:
        print(f"unzip {data_path}")


def web_download_un7z(
    url, zip_path, filename="", force_download=False, verbose=True, **kwargs
):
    import py7zr

    web_download(url, zip_path, filename, force_download)
    # assume that path/to/abc.zip consists path/to/abc
    data_path = zip_path[:-3]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    # data_root = os.path.dirname(zip_path)
    with py7zr.SevenZipFile(zip_path, mode="r") as z:
        z.extractall()
    if verbose:
        print(f"un7z {data_path}")


def web_download_untar(
    url, tar_path, filename="", force_download=False, verbose=True, **kwargs
):
    web_download(url, tar_path, filename, force_download)
    # assume that path/to/abc.tar consists path/to/abc
    data_path = tar_path[:-4]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    data_root = os.path.dirname(tar_path)
    with tarfile.open(tar_path) as tar:

        def is_within_directory(directory, target):

            abs_directory = os.path.abspath(directory)
            abs_target = os.path.abspath(target)

            prefix = os.path.commonprefix([abs_directory, abs_target])

            return prefix == abs_directory

        def safe_extract(tar, path=".", members=None, *, numeric_owner=False):

            for member in tar.getmembers():
                member_path = os.path.join(path, member.name)
                if not is_within_directory(path, member_path):
                    raise Exception("Attempted Path Traversal in Tar File")

            tar.extractall(path, members, numeric_owner=numeric_owner)

        safe_extract(tar, data_root)
    if verbose:
        print(f"decompress {tar_path}")


def web_download_ungzip(
    url, gzip_path, filename="", force_download=False, verbose=True, **kwargs
):
    web_download(url, gzip_path, filename, force_download)
    # assume that path/to/abc.gzip consists path/to/abc
    data_path = gzip_path[:-3]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    with gzip.open(gzip_path, "rb") as fi:
        with open(data_path, "wb") as fo:
            shutil.copyfileobj(fi, fo)
    print(f"decompress {gzip_path}")


def gdrive_download_un7z(
    file_id, zip_path, filename="", force_download=False, verbose=True, **kwargs
):
    import py7zr

    gdrive_download(file_id, zip_path, filename, force_download)
    # assume that path/to/abc.zip consists path/to/abc
    data_path = zip_path[:-3]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    # data_root = os.path.dirname(zip_path)
    with py7zr.SevenZipFile(zip_path, mode="r") as z:
        z.extractall(path=data_path)
    if verbose:
        print(f"un7z {data_path}")


def gdrive_download_untar(
    file_id, local_path, filename="", force_download=False, verbose=True, **kwargs
):
    gdrive_download(file_id, local_path, filename, force_download)
    # assume that path/to/abc.tar consists path/to/abc
    data_path = local_path[: local_path.find(".tar")]
    if (not force_download) and os.path.exists(data_path):
        if verbose:
            print(f"[{filename}] is already extracted at {data_path}")
        return None
    data_root = os.path.dirname(local_path)
    with tarfile.open(local_path) as tar:

        def is_within_directory(directory, target):

            abs_directory = os.path.abspath(directory)
            abs_target = os.path.abspath(target)

            prefix = os.path.commonprefix([abs_directory, abs_target])

            return prefix == abs_directory

        def safe_extract(tar, path=".", members=None, *, numeric_owner=False):

            for member in tar.getmembers():
                member_path = os.path.join(path, member.name)
                if not is_within_directory(path, member_path):
                    raise Exception("Attempted Path Traversal in Tar File")

            tar.extractall(path, members, numeric_owner=numeric_owner)

        safe_extract(tar, data_root)
    if verbose:
        print(f"decompress {local_path}")


def gdrive_download(
    file_id,
    local_path,
    filename="",
    force_download=False,
    verify=True,
    verbose=True,
    **kwargs,
):
    import gdown

    if filename == "":
        filename = os.path.basename(local_path)
    if (not force_download) and os.path.exists(local_path):
        if verbose:
            print(f"[ekorpkit] [{filename}] is already downloaded at {local_path}")
        return None

    if os.path.dirname(local_path):
        os.makedirs(os.path.dirname(local_path), exist_ok=True)

    gdown.download(
        id=file_id,
        output=local_path,
        quiet=not verbose,
        fuzzy=True,
        verify=verify,
    )
    if verbose:
        print(f"download {filename}")



import collections
import logging
import numpy as np
from typing import Iterator, Optional, Union
from tqdm.auto import tqdm
from scipy.special import digamma
from .base import Trainer
from ..utils.trie import Trie


log = logging.getLogger(__name__)


class UnigramTrainer(Trainer):
    """
    Trainer capable of training a Unigram model
    Args:
        vocab_size (:obj:`int`):
            The size of the final vocabulary, including all tokens and alphabet.
        show_progress (:obj:`bool`):
            Whether to show progress bars while training.
        special_tokens (:obj:`List[Union[str, AddedToken]]`):
            A list of special tokens the model should know of.
        initial_alphabet (:obj:`List[str]`):
            A list of characters to include in the initial alphabet, even
            if not seen in the training dataset.
            If the strings contain more than one character, only the first one
            is kept.
        shrinking_factor (:obj:`float`):
            The shrinking factor used at each step of the training to prune the
            vocabulary.
        unk_token (:obj:`str`):
            The token used for out-of-vocabulary tokens.
        max_piece_length (:obj:`int`):
            The maximum length of a given token.
        n_sub_iterations (:obj:`int`):
            The number of iterations of the EM algorithm to perform before
            pruning the vocabulary.
    """

    def __init__(
        self,
        vocab_size=8000,
        show_progress=True,
        special_tokens=[],
        initial_alphabet=[],
        shrinking_factor=0.75,
        unk_token=None,
        max_piece_length=16,
        n_sub_iterations=5,
        max_rounds=5,
        delta=0.01,
        min_frequency: int = 2,
        verbose=False,
    ):
        self.vocab_size = vocab_size
        self.show_progress = show_progress
        self.special_tokens = special_tokens
        self.initial_alphabet = initial_alphabet
        self.shrinking_factor = shrinking_factor
        self.unk_token = unk_token
        self.max_piece_length = max_piece_length
        self.n_sub_iterations = n_sub_iterations
        self.max_rounds = max_rounds
        self.delta = delta
        self.min_frequency = min_frequency
        self.verbose = verbose

        super().__init__()

    def get_word_freqs(self, words):
        word_freqs = collections.Counter(words)
        # Remove words that are too rare
        word_freqs = {
            word: freq
            for word, freq in word_freqs.items()
            if freq >= self.min_frequency
        }
        return word_freqs

    def initialize_subwords(self, word_freqs):
        character_freqs = collections.defaultdict(int)
        subwords_freqs = collections.defaultdict(int)
        for word, freq in word_freqs.items():
            # word = self.whitespace_token + word
            for i in range(len(word)):
                character_freqs[word[i]] += freq
                # Loop through the subwords of length at least 2
                for j in range(i + 2, len(word) + 1):
                    subwords_freqs[word[i:j]] += freq
        # Sort subwords by frequency
        sorted_subwords = sorted(
            subwords_freqs.items(), key=lambda x: x[1], reverse=True
        )

        return sorted_subwords, character_freqs

    def initialize_vocab(self, words):
        word_freqs = self.get_word_freqs(words)
        sorted_subwords, characters = self.initialize_subwords(word_freqs)

        alphabet = {
            char: self.min_frequency
            for char in self.initial_alphabet
            if char not in characters
        }
        characters.update(alphabet)
        tokens = list(characters.items()) + sorted_subwords
        tokens = {token: freq for token, freq in tokens}
        tokens = collections.Counter(tokens)
        return tokens, characters

    def initialize_trie(self, tokens):
        trie = Trie()
        norm = sum(list(tokens.values()))
        logsum = digamma(norm)

        maxlen = 0
        for tok, val in tokens.items():
            trie.add(tok, digamma(val) - logsum)
            maxlen = max(maxlen, len(tok))

        return trie, maxlen

    def forward_step(self, text, trie):
        N = len(text)

        # d[i] contains the maximum log_prob of any tokenization
        # of text[:i], initialized to 0 (i.e. log(0)=-infty)
        d = [-np.inf] * (N + 1)

        # p[i] (stands for parent) contains the number of characters of
        # the final token in the most likely sequence that ends at index i
        p = [None] * (N + 1)
        d[0] = 0

        for i in range(1, N + 1):

            # find all possible final words. Have to look back
            # a distance set by the length of the longest token
            for j in range(max(i - self.maxlen, 0), i):

                final_token = text[j:i]
                final_value = trie.get_value(final_token)

                # if the current ending word has a higher log-probability,
                # save that value and store the word (i.e. # chars to backtrack)
                if final_value and d[j] + final_value > d[i]:
                    d[i] = d[j] + final_value
                    p[i] = len(final_token)
            if p[i] is None:
                raise ValueError(f"Encountered unknown token '{text[i-1]}'.")

        loss = d[-1]
        return loss, p

    def backward_step(self, text, p):
        idx = len(p)
        tokenization = []
        while idx > 1:
            # move back the number of steps p tells you to
            next_idx = idx - p[idx - 1]

            # extract the final token
            tok = text[next_idx - 1 : idx - 1]
            tokenization.append(tok)

            idx = next_idx
        tokenization = list(reversed(tokenization))
        return tokenization

    def E_step(self, tokenization, trie):
        # get the new token counts based on updated tokenization
        counts = collections.Counter(tokenization)
        norm = sum(list(counts.values()))

        # we are returning the log probabilties here (alpha=0 prior)
        logsum = digamma(norm)
        for k, v in counts.items():
            counts[k] = digamma(v) - logsum

        for k, v in counts.items():
            trie.set_value(k, v)
        return trie

    def M_step(self, text, trie):
        loss, p = self.forward_step(text, trie)
        tokenization = self.backward_step(text, p)
        return tokenization, loss

    def EM_step(self, text, tokenization, trie):
        trie = self.E_step(tokenization, trie)
        tokenization, loss = self.M_step(text, trie)
        return loss, tokenization, trie

    def EM_round(self, text, tokens, delta=0.01, n_iterations=10):
        tokenization, old_loss = self.M_step(text, self.trie)
        for step in range(n_iterations):
            print(f"EM iter {step}: ", end="")
            loss, tokenization, trie = self.EM_step(text, tokenization, self.trie)
            print(f"Loss={loss:.2f}")
            if abs(old_loss - loss) < delta:
                break
            old_loss = loss

    def prune_tokens(self, tokens, characters, vocab_size, shrinking_factor=None):
        """Tokens are passed by reference and modified in place.
        Returns:
            True: to indicate to caller that more rounds are needed
            False: to indicate we successfully hit the target vocab size
            ValueError: if the vocab size cannot be reached."""
        sorted_tokens = tokens.most_common()
        N = len(sorted_tokens)
        if shrinking_factor is None:
            shrinking_factor = self.shrinking_factor
        n_trim = int((1 - shrinking_factor) * N)
        for i in reversed(range(N)):
            if N <= vocab_size:
                return False
            if n_trim <= 0:
                return True
            tok = sorted_tokens[i][0]
            if tok not in characters:
                self.trie.set_value(
                    tok, 0
                )  # we need to delete it from the trie (that sticks around)
                tokens.pop(
                    tok
                )  # also need to delete from tokens, so the next round doesn't see it
                n_trim -= 1
                N -= 1
        if n_trim > 0:
            raise ValueError(
                "Could not reduce tokens further. Please increase vocab size"
            )
        return False

    def fit(self, texts):
        """To turn off pruning, just set max_rounds=1"""
        words = []
        iterator = tqdm(texts) if self.show_progress else texts
        for text in iterator:
            words += self.pre_tokenize(text)
        text = "".join(words)
        tokens, characters = self.initialize_vocab(words)
        vocab_size = self.vocab_size

        if vocab_size > len(tokens):
            raise ValueError(
                f"Vocab size is larger than the availble number of tokens {len(tokens)}."
            )
        self.trie, self.maxlen = self.initialize_trie(tokens)
        for i in range(1, self.max_rounds + 1):
            print(f"--- Round {i}. Vocab size: {len(tokens)} ---")
            self.EM_round(text, tokens, self.delta, self.n_sub_iterations)
            if not self.prune_tokens(tokens, characters, vocab_size):
                break
        self.vocab_size = len(tokens)
        return tokens

    def train(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        length: Optional[int] = None,
    ):
        """Train the model using the given iterator"""

        return self.fit(iterator)



import inspect
import warnings
import numpy as np
import PIL
import torch
from diffusers import (
    AutoencoderKL,
    DDIMScheduler,
    DiffusionPipeline,
    LMSDiscreteScheduler,
    PNDMScheduler,
    UNet2DConditionModel,
)
from diffusers.pipelines.stable_diffusion import StableDiffusionSafetyChecker
from transformers import CLIPFeatureExtractor, CLIPTextModel, CLIPTokenizer
from typing import List, Optional, Union


def scale_image(image, max_pixels):
    """Scale image to have at most `max_pixels` pixels."""
    w, h = image.size
    scale = np.sqrt(max_pixels / (w * h))
    if scale < 1.0:
        image = image.resize(
            (int(w * scale), int(h * scale)), resample=PIL.Image.LANCZOS
        )
    return image


def preprocess(image, max_pixels=262144):
    # remove alpha channel

    image = image.convert("RGB")

    # if image is bigger than total number of pixels, scale it down
    image = scale_image(image, max_pixels=max_pixels)

    w, h = image.size
    w, h = map(lambda x: x - x % 64, (w, h))  # resize to integer multiple of 32
    image = image.resize((w, h), resample=PIL.Image.LANCZOS)
    image = np.array(image).astype(np.float32) / 255.0
    image = image[None].transpose(0, 3, 1, 2)
    image = torch.from_numpy(image)

    return 2.0 * image - 1.0


class StableDiffusionPipeline(DiffusionPipeline):
    def __init__(
        self,
        vae: AutoencoderKL,
        text_encoder: CLIPTextModel,
        tokenizer: CLIPTokenizer,
        unet: UNet2DConditionModel,
        scheduler: Union[DDIMScheduler, PNDMScheduler, LMSDiscreteScheduler],
        safety_checker: StableDiffusionSafetyChecker,
        feature_extractor: CLIPFeatureExtractor,
    ):
        super().__init__()
        scheduler = scheduler.set_format("pt")
        self.register_modules(
            vae=vae,
            text_encoder=text_encoder,
            tokenizer=tokenizer,
            unet=unet,
            scheduler=scheduler,
            safety_checker=safety_checker,
            feature_extractor=feature_extractor,
        )

    @torch.no_grad()
    def __call__(
        self,
        prompt: Union[str, List[str]],
        height: Optional[int] = 512,
        width: Optional[int] = 512,
        init_image: Optional[Union[torch.FloatTensor, PIL.Image.Image]] = None,
        init_max_pixels: int = 262144,
        init_strength: float = 0.8,
        num_inference_steps: Optional[int] = 50,
        guidance_scale: Optional[float] = 7.5,
        eta: Optional[float] = 0.0,
        generator: Optional[torch.Generator] = None,
        output_type: Optional[str] = "pil",
        **kwargs,
    ):
        if "torch_device" in kwargs:
            device = kwargs.pop("torch_device")
            warnings.warn(" Consider using `pipe.to(torch_device)` instead.")

            # Set device as before (to be removed in 0.3.0)
            if device is None:
                device = "cuda" if torch.cuda.is_available() else "cpu"
            self.to(device)

        if isinstance(prompt, str):
            batch_size = 1
        elif isinstance(prompt, list):
            batch_size = len(prompt)
        else:
            raise ValueError(
                f"`prompt` has to be of type `str` or `list` but is {type(prompt)}"
            )

        if init_image is not None and height % 8 != 0 or width % 8 != 0:
            raise ValueError(
                f"`height` and `width` have to be divisible by 8 but are {height} and {width}."
            )

        if init_image is not None:
            if init_strength < 0 or init_strength > 1:
                raise ValueError(
                    f"The value of strength should in [0.0, 1.0] but is {init_strength}"
                )

        # get prompt text embeddings
        text_input = self.tokenizer(
            prompt,
            padding="max_length",
            max_length=self.tokenizer.model_max_length,
            truncation=True,
            return_tensors="pt",
        )
        text_embeddings = self.text_encoder(text_input.input_ids.to(self.device))[0]

        # here `guidance_scale` is defined analog to the guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf . `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0
        # get unconditional embeddings for classifier free guidance
        if do_classifier_free_guidance:
            max_length = text_input.input_ids.shape[-1]
            uncond_input = self.tokenizer(
                [""] * batch_size,
                padding="max_length",
                max_length=max_length,
                return_tensors="pt",
            )
            uncond_embeddings = self.text_encoder(
                uncond_input.input_ids.to(self.device)
            )[0]

            # For classifier free guidance, we need to do two forward passes.
            # Here we concatenate the unconditional and text embeddings into a single batch
            # to avoid doing two forward passes
            text_embeddings = torch.cat([uncond_embeddings, text_embeddings])

        # set timesteps
        accepts_offset = "offset" in set(
            inspect.signature(self.scheduler.set_timesteps).parameters.keys()
        )
        extra_set_kwargs = {}
        offset = 0
        if accepts_offset:
            offset = 1
            extra_set_kwargs["offset"] = 1

        self.scheduler.set_timesteps(num_inference_steps, **extra_set_kwargs)

        if init_image is not None:
            if not isinstance(init_image, torch.FloatTensor):
                init_image = preprocess(init_image, max_pixels=init_max_pixels)

            # encode the init image into latents and scale the latents
            init_latents = self.vae.encode(init_image.to(self.device)).sample()
            init_latents = 0.18215 * init_latents

            # expand init_latents for batch_size
            init_latents = torch.cat([init_latents] * batch_size)

            # get the original timestep using init_timestep
            init_timestep = int(num_inference_steps * init_strength) + offset
            init_timestep = min(init_timestep, num_inference_steps)
            if isinstance(self.scheduler, LMSDiscreteScheduler):
                timesteps = torch.tensor(
                    [num_inference_steps - init_timestep] * batch_size,
                    dtype=torch.long,
                    device=self.device,
                )
            else:
                timesteps = self.scheduler.timesteps[-init_timestep]
                timesteps = torch.tensor(
                    [timesteps] * batch_size, dtype=torch.long, device=self.device
                )

            # add noise to latents using the timesteps
            noise = torch.randn(
                init_latents.shape, generator=generator, device=self.device
            )
            init_latents = self.scheduler.add_noise(init_latents, noise, timesteps).to(
                self.device
            )
            latents = init_latents

        else:
            # get the intial random noise
            latents = torch.randn(
                (batch_size, self.unet.in_channels, height // 8, width // 8),
                generator=generator,
                device=self.device,
            )

            # if we use LMSDiscreteScheduler, let's make sure latents are mulitplied by sigmas
            if isinstance(self.scheduler, LMSDiscreteScheduler):
                latents = latents * self.scheduler.sigmas[0]

        # prepare extra kwargs for the scheduler step, since not all schedulers have the same signature
        # eta (η) is only used with the DDIMScheduler, it will be ignored for other schedulers.
        # eta corresponds to η in DDIM paper: https://arxiv.org/abs/2010.02502
        # and should be between [0, 1]
        accepts_eta = "eta" in set(
            inspect.signature(self.scheduler.step).parameters.keys()
        )
        extra_step_kwargs = {}
        if accepts_eta:
            extra_step_kwargs["eta"] = eta

        t_start = max(num_inference_steps - init_timestep + offset, 0)
        for i, t in enumerate(self.scheduler.timesteps[t_start:]):
            t_index = t_start + i
            # expand the latents if we are doing classifier free guidance
            latent_model_input = (
                torch.cat([latents] * 2) if do_classifier_free_guidance else latents
            )
            # if we use LMSDiscreteScheduler, let's make sure latents are mulitplied by sigmas
            if isinstance(self.scheduler, LMSDiscreteScheduler):
                sigma = self.scheduler.sigmas[t_index]
                # the model input needs to be scaled to match the continuous ODE formulation in K-LMS
                latent_model_input = latent_model_input / ((sigma**2 + 1) ** 0.5)
                latent_model_input = latent_model_input.to(self.unet.dtype)
                t = t.to(self.unet.dtype)

            # predict the noise residual
            noise_pred = self.unet(
                latent_model_input, t, encoder_hidden_states=text_embeddings
            )["sample"]

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + guidance_scale * (
                    noise_pred_text - noise_pred_uncond
                )

            # compute the previous noisy sample x_t -> x_t-1
            if isinstance(self.scheduler, LMSDiscreteScheduler):
                latents = self.scheduler.step(
                    noise_pred, t_index, latents, **extra_step_kwargs
                )["prev_sample"]
            else:
                latents = self.scheduler.step(
                    noise_pred, t, latents, **extra_step_kwargs
                )["prev_sample"]

        # scale and decode the image latents with vae
        latents = 1 / 0.18215 * latents
        image = self.vae.decode(latents)

        image = (image / 2 + 0.5).clamp(0, 1)
        image = image.cpu().permute(0, 2, 3, 1).numpy()

        # run safety checker
        safety_cheker_input = self.feature_extractor(
            self.numpy_to_pil(image), return_tensors="pt"
        ).to(self.device)
        image, has_nsfw_concept = self.safety_checker(
            images=image, clip_input=safety_cheker_input.pixel_values
        )

        if output_type == "pil":
            image = self.numpy_to_pil(image)

        return {"sample": image, "nsfw_content_detected": has_nsfw_concept}



import os
import json
import logging
import matplotlib.pyplot as plt
from tqdm.auto import tqdm
from typing import Iterator, Optional, Union
from ekorpkit.visualize.base import get_plot_font
from .base import Model
from ..trainers.branching import BranchingEntropyTrainer
from ..utils.trie import Trie, entropy


log = logging.getLogger(__name__)


class BranchingEntropy(Model):
    def __init__(
        self,
        vocab=None,
        branching_threshold=0.2,
        whitespace_token="▁",
        whitespace_token_as_prefix=True,
        verbose=False,
        **kwargs,
    ):
        super().__init__(vocab)
        self.branching_threshold = branching_threshold
        self.whitespace_token = whitespace_token
        self.whitespace_token_as_prefix = whitespace_token_as_prefix
        self.verbose = verbose

    def initialize_vocab(self, vocab, **kwargs):
        self.vocab = {}
        self.token2id = {}
        self.id2token = {}
        self.fwd_trie = None
        self.bwd_trie = None
        self.max_piece_length = None
        if vocab:
            self.vocab = vocab
            self.fwd_trie, self.max_piece_length = self.initialize_trie(
                vocab, "forward"
            )
            self.bwd_trie, _ = self.initialize_trie(vocab, "backward")

    def initialize_trie(self, tokens, direction="forward"):
        trie = Trie(direction=direction)

        maxlen = 0
        for tok, val in tqdm(tokens.items(), desc=f"Building {direction} trie"):
            trie.add(tok, val)
            maxlen = max(maxlen, len(tok))

        return trie, maxlen

    def get_entropy(self, word, direction="forward"):
        if direction == "forward":
            _trie = self.fwd_trie
        else:
            _trie = self.bwd_trie
        return entropy(_trie, word)

    def score(self, word, direction="forward"):
        return self.get_entropy(word, direction=direction)

    def find_local_entropy(self, word, direction="forward"):
        # get the local entropy and the difference in entropy

        entropies = []
        for i in range(1, len(word) + 1):
            if direction == "forward":
                if word.startswith(self.whitespace_token):
                    subword = word[:i]
                else:
                    subword = self.whitespace_token + word[:i]
            else:
                if word.endswith(self.whitespace_token):
                    subword = word[-i:]
                else:
                    subword = word[-i:] + self.whitespace_token
            _score = self.get_entropy(subword, direction=direction)
            entropies.append(_score)
            if self.verbose:
                print(subword, _score)
        if direction == "backward":
            entropies = entropies[::-1]

        # # get diffs
        if direction == "forward":
            diffs = [0.0] + [
                (entropies[i] - entropies[i - 1]) for i in range(1, len(entropies))
            ]
        else:
            diffs = [
                (entropies[i + 1] - entropies[i]) for i in range(0, len(entropies) - 1)
            ] + [0.0]

        return list(zip(word, entropies, diffs))

    # plot entropies
    def plot_local_entropy(self, word, direction="forward", figsize=(12, 5)):
        get_plot_font()

        results = self.find_local_entropy(word, direction=direction)
        chars, entropies, diffs = zip(*results)
        plt.figure(figsize=figsize)
        plt.plot(entropies, label="entropy", marker="o")
        plt.xticks(range(len(chars)), chars)
        plt.legend(loc="upper left")

        # plot diffs on the right y-axis
        # plt.twinx()
        # plt.plot(diffs, label="diffs", color="red", linestyle="--", marker="o")
        # plt.legend(loc="upper right")
        plt.show()

    def tokenize_word(self, word, direction="forward"):
        # if there is a spike in entropy, then we should segment
        # Here the spike means that there is a sudden increase in entropy followed by a decrease.
        # We can use the difference in entropy to detect the spike.

        # if word.startswith(self.whitespace_token):
        #     word = word[len(self.whitespace_token) :]
        # get the local entropy and the difference in entropy
        results = self.find_local_entropy(word, direction=direction)
        _, _, diffs = zip(*results)

        # get the spikes
        spikes = []
        for i in range(1, len(diffs) - 1):
            if diffs[i] > self.branching_threshold and diffs[i + 1] < 0:
                spikes.append(i)

        # segment the word
        segments = []
        start = 0
        for spike in spikes:
            segments.append(word[start : spike + 1])
            start = spike + 1
        if start < len(word):
            segments.append(word[start:])
        # if self.whitespace_token_as_prefix and len(segments) > 0:
        #     segments[0] = self.whitespace_token + segments[0]
        return tuple(segments)

    def tokenize(
        self, sequence, direction="forward", flatten=True, branching_threshold=None
    ):
        if branching_threshold is not None:
            self.branching_threshold = branching_threshold
        segments = []
        words = self.pre_tokenize(sequence)
        for word in words:
            segments.append(self.tokenize_word(word, direction=direction))
        if flatten:
            segments = [seg for word in segments for seg in word]
        return segments

    def tokenize_texts(self, texts, direction="forward"):
        return [self.tokenize(text, direction=direction) for text in texts]

    def naive_segment(self, text, direction="forward"):
        words = []

        _start, _pos = 0, 0
        # iterate over the text until we reach the end
        while _pos < len(text):
            _sentencepiece = text[_pos : _pos + self.max_piece_length]
            # print(_start, _pos, _sentencepiece)
            if len(_sentencepiece) < 1:
                break
            results = self.find_local_entropy(_sentencepiece, direction=direction)
            _, entropies, _ = zip(*results)

            if entropies[0] == 0:
                if _pos == len(text) - 1:
                    words.append(text[_start : _pos + 1])
                    _start = _pos + 1
                    break
                _pos += 1
            else:
                if _pos > _start:
                    words.append(text[_start:_pos])
                    _start = _pos
                    _pos += 1
                if len(entropies) > 1:
                    _pos += 1
                    for i in range(1, len(entropies)):
                        if entropies[i] == 0:
                            words.append(text[_start : _start + i])
                            # print(_start, words)
                            _start += i
                            _pos = _start
                            break
                        elif _pos == len(text) - 1:
                            words.append(text[_start : _pos + 1])
                        _pos += 1
                else:
                    if _pos == len(text) - 1:
                        words.append(text[_start : _pos + 1])
                    _pos += 1

        return words

    def train_from_iterator(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        trainer: BranchingEntropyTrainer = None,
        length: Optional[int] = None,
    ):
        """Train the model using the given iterator"""

        trainer.normalizer = self.normalizer
        trainer.pre_tokenizer = self.pre_tokenizer
        vocab = trainer.train(iterator, length=length)
        self.initialize_vocab(vocab)

    @classmethod
    def from_file(cls, vocab, **kwargs):
        """
        Instantiate a BranchingEntropy model from the given files.

        This method is roughly equivalent to doing::

           vocab = BranchingEntropy.read_file(vocab_filename)
           be = BranchingEntropy(vocab)

        Args:
            vocab (:obj:`str`):
                The path to a :obj:`vocab.json` file

        Returns:
            :class:`~tokenizers.models.BranchingEntropy`: An instance of BranchingEntropy loaded from these files
        """
        vocab = cls.read_file(vocab)
        return cls(vocab, **kwargs)

    @staticmethod
    def read_file(vocab):
        """
        Read a :obj:`vocab.json` file

        This method provides a way to read and parse the content of these files,
        returning the relevant data structures. If you want to instantiate some BPE models
        from memory, this method gives you the expected input from the standard files.

        Args:
            vocab (:obj:`str`):
                The path to a :obj:`vocab.json` file

        Returns:
            A :obj:`Tuple` with the vocab and the merges:
                The vocabulary and merges loaded into memory
        """
        with open(vocab, "r") as f:
            vocab = json.load(f)
        return vocab

    def save(self, folder, prefix=None, pretty: bool = False):
        """
        Save the current model

        Save the current model in the given folder, using the given prefix for the various
        files that will get created.
        Any file with the same name that already exists in this folder will be overwritten.

        Args:
            folder (:obj:`str`):
                The path to the target folder in which to save the various files

            prefix (:obj:`str`, `optional`):
                An optional prefix, used to prefix each file name

        Returns:
            :obj:`List[str]`: The list of saved files
        """
        if prefix is not None:
            folder = os.path.join(folder, prefix)
        vocab_filename = os.path.join(folder, "vocab.json")
        if not os.path.exists(folder):
            os.makedirs(folder)
        indent = 2 if pretty else None
        json.dump(
            self.vocab, open(vocab_filename, "w"), indent=indent, ensure_ascii=False
        )
        return [vocab_filename]



import json
import os
import logging
from typing import Iterator, List, Optional, Union
from .base import Model
from ..trainers.bpe import BpeTrainer


log = logging.getLogger(__name__)


class BPE(Model):
    """
    An implementation of the BPE (Byte-Pair Encoding) algorithm

    Args:
        vocab (:obj:`Dict[str, int]`, `optional`):
            A dictionnary of string keys and their ids :obj:`{"am": 0,...}`

        merges (:obj:`List[Tuple[str, str]]`, `optional`):
            A list of pairs of tokens (:obj:`Tuple[str, str]`) :obj:`[("a", "b"),...]`

        cache_capacity (:obj:`int`, `optional`):
            The number of words that the BPE cache can contain. The cache allows
            to speed-up the process by keeping the result of the merge operations
            for a number of words.

        dropout (:obj:`float`, `optional`):
            A float between 0 and 1 that represents the BPE dropout to use.

        unk_token (:obj:`str`, `optional`):
            The unknown token to be used by the model.

        continuing_subword_prefix (:obj:`str`, `optional`):
            The prefix to attach to subword units that don't represent a beginning of word.

        end_of_word_suffix (:obj:`str`, `optional`):
            The suffix to attach to subword units that represent an end of word.

        fuse_unk (:obj:`bool`, `optional`):
            Whether to fuse any subsequent unknown tokens into a single one
    """

    def __init__(
        self,
        vocab=None,
        merges=None,
        cache_capacity=None,
        dropout=None,
        unk_token="</u>",
        continuing_subword_prefix=None,
        end_of_word_suffix="</w>",
        fuse_unk=None,
    ):
        super().__init__(vocab=vocab, merges=merges)
        self.cache_capacity = cache_capacity
        self.dropout = dropout
        self.unk_token = unk_token
        self.continuing_subword_prefix = continuing_subword_prefix
        self.end_of_word_suffix = end_of_word_suffix
        self.fuse_unk = fuse_unk

    def initialize_vocab(self, vocab, merges, **kwargs):
        self.vocab = {}
        self.merges = {}
        self.token2id = {}
        self.id2token = {}
        if vocab:
            self.vocab = vocab
            self.merges = merges
            self.token2id = {token: i for i, token in enumerate(vocab)}
            self.id2token = {v: k for k, v in self.token2id.items()}

    @classmethod
    def from_file(cls, vocab, merges, **kwargs):
        """
        Instantiate a BPE model from the given files.

        This method is roughly equivalent to doing::

           vocab, merges = BPE.read_file(vocab_filename, merges_filename)
           bpe = BPE(vocab, merges)

        If you don't need to keep the :obj:`vocab, merges` values lying around,
        this method is more optimized than manually calling
        :meth:`~tokenizers.models.BPE.read_file` to initialize a :class:`~tokenizers.models.BPE`

        Args:
            vocab (:obj:`str`):
                The path to a :obj:`vocab.json` file

            merges (:obj:`str`):
                The path to a :obj:`merges.txt` file

        Returns:
            :class:`~tokenizers.models.BPE`: An instance of BPE loaded from these files
        """
        vocab, merges = cls.read_file(vocab, merges)
        return cls(vocab, merges, **kwargs)

    @staticmethod
    def read_file(vocab, merges):
        """
        Read a :obj:`vocab.json` and a :obj:`merges.txt` files

        This method provides a way to read and parse the content of these files,
        returning the relevant data structures. If you want to instantiate some BPE models
        from memory, this method gives you the expected input from the standard files.

        Args:
            vocab (:obj:`str`):
                The path to a :obj:`vocab.json` file

            merges (:obj:`str`):
                The path to a :obj:`merges.txt` file

        Returns:
            A :obj:`Tuple` with the vocab and the merges:
                The vocabulary and merges loaded into memory
        """
        with open(vocab, "r") as f:
            vocab = json.load(f)
        with open(merges, "r") as f:
            merges = [tuple(line.rstrip().split()) for line in f]
            merges = {merge: merge[0] + merge[1] for merge in merges}
        return vocab, merges

    def save(self, folder, prefix=None, pretty: bool = False):
        """
        Save the current model

        Save the current model in the given folder, using the given prefix for the various
        files that will get created.
        Any file with the same name that already exists in this folder will be overwritten.

        Args:
            folder (:obj:`str`):
                The path to the target folder in which to save the various files

            prefix (:obj:`str`, `optional`):
                An optional prefix, used to prefix each file name

        Returns:
            :obj:`List[str]`: The list of saved files
        """
        if prefix is not None:
            folder = os.path.join(folder, prefix)
        vocab_filename = os.path.join(folder, "vocab.json")
        merges_filename = os.path.join(folder, "merges.txt")
        if not os.path.exists(folder):
            os.makedirs(folder)
        indent = 2 if pretty else None
        json.dump(self.vocab, open(vocab_filename, "w"), indent=indent)
        with open(merges_filename, "w") as f:
            for merge in self.merges:
                f.write(f"{merge[0]} {merge[1]}" + os.linesep)
        print(f"Model files saved in {folder}")
        return [vocab_filename, merges_filename]

    def decode(self, ids: List[int], skip_special_tokens: Optional[bool] = True) -> str:
        """Decode the given list of ids to a string sequence

        Args:
            ids: List[unsigned int]:
                A list of ids to be decoded

            skip_special_tokens: (`optional`) boolean:
                Whether to remove all the special tokens from the output string

        Returns:
            The decoded string
        """
        if ids is None:
            raise ValueError("None input is not valid. Should be a list of integers.")

        tokens = [self.id_to_token(id) for id in ids]
        # if skip_special_tokens:
        #     tokens = [token for token in tokens if token not in self.special_tokens]

        return "".join(tokens).replace(self.end_of_word_suffix, " ").strip()

    def tokenize(self, sequence):
        """
        Tokenize a sequence

        Args:
            sequence (:obj:`str`):
                A sequence to tokenize

        Returns:
            A :obj:`List` of :class:`~tokenizers.Token`: The generated tokens
        """
        pre_tokenized_text = self.pre_tokenize(sequence)
        splits = [
            [char for char in word] + [self.end_of_word_suffix]
            for word in pre_tokenized_text
        ]
        for pair, merge in self.merges.items():
            for idx, split in enumerate(splits):
                i = 0
                while i < len(split) - 1:
                    if split[i] == pair[0] and split[i + 1] == pair[1]:
                        split = split[:i] + [merge] + split[i + 2 :]
                    else:
                        i += 1
                splits[idx] = split

        return sum(splits, [])

    def train_from_iterator(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        trainer: BpeTrainer = None,
        length: Optional[int] = None,
    ):
        """Train the model using the given iterator"""

        trainer.normalizer = self.normalizer
        trainer.pre_tokenizer = self.pre_tokenizer
        vocab, merges = trainer.train(iterator, length=length)
        self.initialize_vocab(vocab, merges)



import math
from scipy.special import digamma


def entropy(trie, word):
    leafs = trie.get_leafs(word)
    val = trie.get_value(word)
    logsum = digamma(sum(leafs) + val)
    entropy = 0
    for freq in leafs:
        logprob = digamma(freq) - logsum
        entropy += math.exp(logprob) * logprob
    return -1 * entropy


class Trie:
    def __init__(self, end_symbol="<END>", direction="forward"):
        self.root = {}
        self.end_symbol = end_symbol
        self.direction = direction

    def add(self, word, value):
        if self.direction == "backward":
            # reverse the word
            word = word[::-1]
        node = self.root
        for ch in word:
            if ch not in node:
                node[ch] = {}
            node = node[ch]
        node[self.end_symbol] = value

    def get_value(self, word):
        if self.direction == "backward":
            # reverse the word
            word = word[::-1]
        node = self.root
        for ch in word:
            if ch not in node:
                return 0
            node = node[ch]
        if self.end_symbol not in node:
            return 0
        return node[self.end_symbol]

    def set_value(self, word, value):
        if self.direction == "backward":
            # reverse the word
            word = word[::-1]
        node = self.root
        for ch in word:
            if ch not in node:
                raise ValueError("word not in trie")
            node = node[ch]
        if self.end_symbol not in node:
            raise ValueError("word not in trie")
        node[self.end_symbol] = value

    def get_children(self, word):
        if self.direction == "backward":
            # reverse the word
            word = word[::-1]
        node = self.root
        for ch in word:
            if ch not in node:
                return []
            node = node[ch]
        children = node.copy()
        return children

    def get_values_of_children(self, word):
        children = self.get_children(word)
        values = []
        for child in children:
            if child == self.end_symbol:
                continue
            else:
                if self.end_symbol in children[child]:
                    values.append(children[child][self.end_symbol])
        return values

    def num_children(self, word):
        return len(self.get_children(word))

    def get_leafs(self, word):
        node = self.get_children(word)
        if not node:
            return []
        if self.end_symbol in node:
            _ = node.pop(self.end_symbol)
        return self._get_leafs(node)

    def _get_leafs(self, node):
        if self.end_symbol in node:
            return [node[self.end_symbol]]
        leafs = []
        for child in node:
            leafs += self._get_leafs(node[child])
        return leafs



import logging
import collections
from typing import Iterator, Optional, Union
from tqdm.auto import tqdm
from .base import Trainer
from ..utils.trie import Trie


log = logging.getLogger(__name__)


class BranchingEntropyTrainer(Trainer):
    def __init__(
        self,
        whitespace_token="▁",
        max_sentencepiece_length=20,
        min_frequency: int = 2,
        show_progress=True,
        verbose=False,
    ):
        self.whitespace_token = whitespace_token
        self.max_sentencepiece_length = max_sentencepiece_length
        self.min_frequency = min_frequency
        self.show_progress = show_progress
        self.verbose = verbose

        self.fwd_trie = None
        self.bwd_trie = None

        super().__init__()

    def get_word_freqs(self, words):
        word_freqs = collections.Counter(words)
        if self.verbose:
            print("Total words:", len(word_freqs))
            print("Top 10 words: {}".format(word_freqs.most_common(10)))
        # Remove words that are too rare
        word_freqs = {
            word: freq
            for word, freq in word_freqs.items()
            if freq >= self.min_frequency
        }
        if self.verbose:
            print("Total words after filtering:", len(word_freqs))
        return word_freqs

    def initialize_subwords(self, word_freqs):
        subwords_freqs = collections.defaultdict(int)

        if self.show_progress:
            iterator = tqdm(word_freqs.items())
        else:
            iterator = word_freqs.items()
        for word, freq in iterator:
            word += self.whitespace_token
            for i in range(len(word)):
                for j in range(
                    i + 1, min(i + self.max_sentencepiece_length, len(word)) + 1
                ):
                    subwords_freqs[word[i:j]] += freq

        # Sort subwords by frequency
        sorted_subwords = sorted(
            subwords_freqs.items(), key=lambda x: x[1], reverse=True
        )
        if self.verbose:
            print("Total subwords:", len(sorted_subwords))
            print("Top 10 subwords: {}".format(sorted_subwords[:10]))

        subwords = collections.Counter(subwords_freqs)
        return subwords

    def initialize_trie(self, tokens, direction="forward"):
        trie = Trie(direction=direction)

        maxlen = 0
        for tok, val in tokens.items():
            trie.add(tok, val)
            maxlen = max(maxlen, len(tok))

        return trie, maxlen

    def fit(self, texts):
        words = []
        iterator = tqdm(texts) if self.show_progress else texts
        for text in iterator:
            words += self.pre_tokenize(text)
        word_freqs = self.get_word_freqs(words)
        vocab = self.initialize_subwords(word_freqs)

        self.fwd_trie, _ = self.initialize_trie(vocab, direction="forward")
        self.bwd_trie, _ = self.initialize_trie(vocab, direction="backward")
        return vocab

    def train(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        length: Optional[int] = None,
    ):
        """Train the model using the given iterator"""

        return self.fit(iterator)

    # def normalize_word(self, word):
    #     # replace all non-alphanumeric characters at the end of the word with a space
    #     word = re.sub(r"[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9]+$", " ", word)
    #     # replace all non-alphanumeric characters at the beginning of the word with a space
    #     word = re.sub(r"^[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9]+", " ", word)
    #     return word.strip()



import logging
import numpy as np
import jax
import jax.numpy as jnp
from functools import partial
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from transformers import CLIPProcessor, FlaxCLIPModel
from flax.jax_utils import replicate
from flax.training.common_utils import shard_prng_key, shard
from PIL import Image
from tqdm.auto import trange
from .base import BaseModel


log = logging.getLogger(__name__)


class DalleMini(BaseModel):
    def __init__(self, root_dir=None, config_name="default", **args):
        cfg = eKonf.compose(f"model/dalle_mini={config_name}")
        cfg = eKonf.merge(cfg, args)
        super().__init__(root_dir=root_dir, **cfg)

        self.model = None
        self.model_params = None
        self.processor = None
        self.vqgan = None
        self.vqgan_params = None
        self.clip = None
        self.clip_params = None
        self.clip_processor = None
        self.devices = None

        if self.autoload:
            self.load()

    def imagine(
        self,
        text_prompts=None,
        batch_name=None,
        batch_num=None,
        show_collage=False,
        **args,
    ):

        """Diffuse the model"""

        log.info("> loading settings...")
        config = self.load_config(
            batch_name=batch_name,
            batch_num=batch_num,
            imagine=args,
        )

        cfg = config.imagine
        text_prompts = text_prompts or eKonf.to_dict(cfg.text_prompts)
        if isinstance(text_prompts, str):
            text_prompts = [text_prompts]
        cfg.text_prompts = text_prompts

        self.sample_imagepaths = []

        tokenized_prompts = self.processor(cfg.text_prompts)
        print(f"Prompts: {cfg.text_prompts}\n")

        batch_dir = self.batch_dir
        num_devices = self.num_devices
        # _num_devices = min(_num_devices, args.num_samples)
        devices = self.devices[:num_devices]
        log.info(f"Using {num_devices} devices")
        log.info(f"Devices: {devices}")
        cfg.num_samples = max(cfg.num_samples // num_devices, 1) * num_devices

        # Keys are passed to the model on each device to generate unique inference per device.
        # create a random key
        log.info(f"Seed used: {self.seed}")
        key = jax.random.PRNGKey(self.seed)
        # Model parameters are replicated on each device for faster inference.
        model_params = replicate(self.model_params, devices=devices)
        vqgan_params = replicate(self.vqgan_params, devices=devices)
        tokenized_prompts = replicate(tokenized_prompts, devices=devices)

        # Model functions are compiled and parallelized to take advantage of multiple devices.
        # model inference
        @partial(
            jax.pmap,
            axis_name="batch",
            static_broadcasted_argnums=(3, 4, 5, 6),
            devices=devices,
        )
        def p_generate(
            tokenized_prompt, key, params, top_k, top_p, temperature, condition_scale
        ):
            return self.model.generate(
                **tokenized_prompt,
                prng_key=key,
                params=params,
                top_k=top_k,
                top_p=top_p,
                temperature=temperature,
                condition_scale=condition_scale,
            )

        # decode image
        @partial(jax.pmap, axis_name="batch", devices=devices)
        def p_decode(indices, params):
            return self.vqgan.decode_code(indices, params=params)

        with elapsed_timer(format_time=True) as elapsed:

            # generate images
            # sample_images = []
            sample_num = 0
            for i in trange(max(cfg.num_samples // num_devices, 1)):
                eKonf.clear_output(wait=True)
                # get a new key
                key, subkey = jax.random.split(key)
                # generate images
                encoded_images = p_generate(
                    tokenized_prompts,
                    shard_prng_key(subkey),
                    model_params,
                    cfg.gen_top_k,
                    cfg.gen_top_p,
                    cfg.temperature,
                    cfg.cond_scale,
                )
                # remove BOS
                encoded_images = encoded_images.sequences[..., 1:]
                # decode images
                decoded_images = p_decode(encoded_images, vqgan_params)
                decoded_images = decoded_images.clip(0.0, 1.0).reshape(
                    (-1, 256, 256, 3)
                )
                for decoded_img in decoded_images:
                    img = Image.fromarray(np.asarray(decoded_img * 255, dtype=np.uint8))
                    eKonf.display(img)
                    # sample_images.append(img)
                    filename = (
                        f"{self.batch_name}({self.batch_num})_{sample_num:04}.png"
                    )
                    img_path = str(batch_dir / filename)
                    img.save(img_path)
                    self.sample_imagepaths.append(img_path)
                    log.info(f"Saved {filename}")
                    sample_num += 1

            eKonf.clear_output(wait=True)
            cfg.num_samples = sample_num
            log.info(" >> elapsed time to diffuse: {}".format(elapsed()))
            print(f"{cfg.num_samples} samples generated to {batch_dir}")
            print(f"text prompts: {text_prompts}")

            if show_collage:
                self.collage(images_or_uris=self.sample_imagepaths)

        config.imagine = cfg
        self.config = config
        results = {
            "image_filepaths": self.sample_imagepaths,
            "config_file": self.save_config(),
            "config": eKonf.to_dict(config),
        }
        return results

    def load_models(self):
        from dalle_mini import DalleBart, DalleBartProcessor
        from vqgan_jax.modeling_flax_vqgan import VQModel

        # check how many devices are available
        log.info(f"Available devices: {jax.local_device_count()}")
        if self.num_devices:
            self.num_devices = min(self.num_devices, jax.local_device_count())
        else:
            self.num_devices = jax.local_device_count()
        self.devices = jax.local_devices()
        log.info(f"Using {self.num_devices} devices")
        log.info(f"Devices: {self.devices}")

        # Load dalle-mini
        self.model, self.model_params = DalleBart.from_pretrained(
            self.model_config.DALLE_MODEL,
            revision=self.model_config.DALLE_COMMIT_ID,
            dtype=jnp.float16,
            _do_init=self.model_config.DALLE_INIT,
        )
        self.processor = DalleBartProcessor.from_pretrained(
            self.model_config.DALLE_MODEL,
            revision=self.model_config.DALLE_COMMIT_ID,
        )
        # Load VQGAN
        self.vqgan, self.vqgan_params = VQModel.from_pretrained(
            self.model_config.VQGAN_REPO,
            revision=self.model_config.VQGAN_COMMIT_ID,
            _do_init=self.model_config.VQGAN_INIT,
        )

    def load_clip_models(self):
        self.clip, self.clip_params = FlaxCLIPModel.from_pretrained(
            self.model_config.CLIP_REPO,
            revision=self.model_config.CLIP_COMMIT_ID,
            dtype=jnp.float16,
            _do_init=self.model_config.CLIP_INIT,
        )
        self.clip_processor = CLIPProcessor.from_pretrained(
            self.model_config.CLIP_REPO,
            revision=self.model_config.CLIP_COMMIT_ID,
        )

    def rank_image_by_clip_score(self, prompts, images):
        """
        Rank an image by its clip score.
        """

        if self.clip is None:
            self.load_clip_models()

        _num_devices = self.num_devices
        _devices = self.devices[:_num_devices]
        log.info(f"Using {_num_devices} devices")
        log.info(f"Devices: {_devices}")

        _clip_params = replicate(self.clip_params, devices=_devices)

        # score images
        @partial(jax.pmap, axis_name="batch", devices=_devices)
        def p_clip(inputs, params):
            logits = self.clip(params=params, **inputs).logits_per_image
            return logits

        # from flax.training.common_utils import shard

        # get clip scores
        clip_inputs = self.clip_processor(
            text=prompts * _num_devices,
            images=images,
            return_tensors="np",
            padding="max_length",
            max_length=77,
            truncation=True,
        ).data
        logits = p_clip(shard(clip_inputs), _clip_params)

        # organize scores per prompt
        p = len(prompts)
        logits = np.asarray([logits[:, i::p, i] for i in range(p)]).squeeze()

        for i, prompt in enumerate(prompts):
            print(f"Prompt: {prompt}\n")
            for idx in logits[i].argsort()[::-1]:
                eKonf.display(images[idx * p + i])
                print(f"Score: {jnp.asarray(logits[i][idx], dtype=jnp.float32):.2f}\n")
            print()



from typing import Iterator, Optional, Union
from tokenizers.pre_tokenizers import PreTokenizer
from tokenizers.normalizers import Normalizer


class Trainer:
    def __init__(
        self,
    ):
        self._normalizer = None
        self._pre_tokenizer = None

    @property
    def normalizer(self) -> Normalizer:
        return self._normalizer

    @normalizer.setter
    def normalizer(self, normalizer: Normalizer):
        self._normalizer = normalizer

    @property
    def pre_tokenizer(self) -> PreTokenizer:
        return self._pre_tokenizer

    @pre_tokenizer.setter
    def pre_tokenizer(self, pre_tokenizer: PreTokenizer):
        self._pre_tokenizer = pre_tokenizer

    def pre_tokenize(self, text):
        text = self.normalizer.normalize_str(text)
        words_with_offsets = self.pre_tokenizer.pre_tokenize_str(text)
        return [word for word, _ in words_with_offsets]

    def initialize_vocab(self, texts):
        pass

    def fit(self, texts, vocab_size=None):
        pass

    def train(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        length: Optional[int] = None,
    ):
        """Train the model using the given iterator"""
        pass



import logging
import random
from omegaconf import OmegaConf
from pathlib import Path
from pydantic import BaseModel, validator
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class BaseBatchConfig(BaseModel):
    output_dir: Path
    batch_name: str
    batch_num: int = None
    random_seed: bool = True
    seed: int = None
    resume_run: bool = False
    resume_latest: bool = False
    config_filesfx = "config.yaml"
    config_dirname = "configs"
    verbose = False

    def __init__(self, **values):
        super().__init__(**values)
        self.init_batch_num()

    def init_batch_num(self):
        if self.batch_num is None:
            num_files = len(list(self.config_dir.glob(self.config_filepattern)))
            if self.resume_latest:
                self.batch_num = num_files - 1
            else:
                self.batch_num = num_files
        if self.verbose:
            log.info(f"Batch name: {self.batch_name}, Batch num: {self.batch_num}")

    @validator("seed")
    def _validate_seed(cls, v, values):
        if values["random_seed"] or v is None or v < 0:
            random.seed()
            seed = random.randint(0, 2**32 - 1)
            if values.get("verbose"):
                log.info(f"Setting seed to {seed}")
            return seed
        return v

    @property
    def batch_dir(self):
        batch_dir = self.output_dir / self.batch_name
        batch_dir.mkdir(parents=True, exist_ok=True)
        return batch_dir

    @property
    def config_dir(self):
        config_dir = self.batch_dir / self.config_dirname
        config_dir.mkdir(parents=True, exist_ok=True)
        return config_dir

    @property
    def file_prefix(self):
        return f"{self.batch_name}({self.batch_num})"

    @property
    def config_filename(self):
        return f"{self.file_prefix}_{self.config_filesfx}"

    @property
    def config_filepattern(self):
        return f"{self.batch_name}(*)_{self.config_filesfx}"

    @property
    def config_filepath(self):
        return self.config_dir / self.config_filename


class BaseConfig:
    _batch_name_ = "demo"
    _batch_num_ = None
    _config_ = None
    _path_ = None
    batch = None

    def __init__(self, root_dir=None, **args):
        self.config = args
        self.verbose = args.get("verbose", False)
        self._init_path(root_dir=root_dir, **args)
        self._init_batch()

    @property
    def autoload(self):
        return self.config.get("autoload", False)

    @property
    def device(self):
        return self.config.get("device", "cpu")

    @property
    def num_devices(self):
        return self.config.get("num_devices")

    @num_devices.setter
    def num_devices(self, num_devices):
        self.config.num_devices = num_devices

    @property
    def version(self):
        return self.config.get("version", "0.0.0")

    @property
    def name(self):
        return self.config.name

    @property
    def path(self):
        return self._path_

    @path.setter
    def path(self, value):
        # self.update("path", value)
        self._path_ = value

    @property
    def config(self):
        return self._config_

    @config.setter
    def config(self, cfg):
        self._config_ = eKonf.to_config(cfg)

    def update(self, key, value, merge=True, force_add=True):
        OmegaConf.update(self._config_, key, value, merge=merge, force_add=force_add)

    @property
    def seed(self):
        return self.batch.seed

    @property
    def root_dir(self):
        return Path(self.path.root)

    @property
    def output_dir(self):
        return Path(self.path.output_dir)

    @property
    def model_config(self):
        if "model" in self.config:
            return self.config.model
        return {}

    @property
    def model_dir(self):
        return Path(self.path.get("model_dir"))

    @property
    def cache_dir(self):
        cache_dir = self.path.get("cache_dir")
        if cache_dir is None:
            cache_dir = self.output_dir / "cache"
            cache_dir.mkdir(parents=True, exist_ok=True)
        return Path(cache_dir)

    @property
    def batch_name(self):
        return self.config.batch.batch_name

    @batch_name.setter
    def batch_name(self, value):
        self.config.batch.batch_name = value

    @property
    def batch_num(self):
        return self.config.batch.batch_num

    @batch_num.setter
    def batch_num(self, value):
        self.config.batch.batch_num = value

    def _init_path(self, path=None, root_dir=None, **kwargs):
        if path is None and self.path is not None:
            path = self.path
            log.info(f"Using existing path: {path.root}")
        if path is None:
            path = self.config.get("path")
            log.info(f"Using config path: {path.root}")
        if path is None:
            path = eKonf.compose("path=_batch_")
            log.info(f"There is no path in the config, using default path: {path.root}")
        if root_dir is not None:
            path.root = root_dir

        if path.verbose:
            eKonf.print(path)
        self.path = path

    def _init_batch(self, verbose=True):
        self.batch = BaseBatchConfig(
            output_dir=self.output_dir, **self.config.batch, verbose=verbose
        )

    def save_config(self, config=None, exclude=["path", "module"], selected=None):
        """Save the batch config"""
        if config is not None:
            self.config = config
        log.info(f"Saving config to {self.batch.config_filepath}")
        cfg = eKonf.to_dict(self.config)
        if selected:
            args = {}
            if isinstance(selected, str):
                selected = [selected]
            for key in selected:
                args[key] = cfg[key]
        else:
            args = cfg
            if exclude:
                if isinstance(exclude, str):
                    exclude = [exclude]
                for key in exclude:
                    args.pop(key, None)
        eKonf.save(args, self.batch.config_filepath)
        return self.batch.config_filename

    def load_config(
        self,
        batch_name=None,
        batch_num=None,
        **args,
    ):
        """Load the config from the batch config file"""
        if batch_name is None:
            batch_name = self.batch_name
        else:
            self.batch_name = batch_name
        self.batch_num = batch_num

        cfg = self.config
        self._init_path()
        self._init_batch(verbose=False)
        if batch_num is not None:
            _path = self.batch.config_filepath
            if _path.is_file():
                log.info(f"Loading config from {_path}")
                batch_cfg = eKonf.load(_path)
                log.info("Merging config with the loaded config")
                cfg = eKonf.merge(cfg, batch_cfg)

        log.info(f"Merging config with args: {args}")
        self.config = eKonf.merge(cfg, args)
        # reinit the batch config to update the config
        self._init_batch()

        return self.config

    def show_config(self, batch_name=None, batch_num=None):
        cfg = self.load_config(batch_name, batch_num)
        eKonf.print(cfg)

    @property
    def module_config(self):
        if "module" in self.config:
            return self.config.module
        return {}

    @property
    def library_dir(self):
        return Path(self.path.library_dir)

    def load_modules(self):
        """Load the modules"""
        if self.module_config.get("modules") is None:
            log.info("No modules to load")
            return
        library_dir = self.library_dir
        for module in self.module_config.modules:
            name = module.name
            libname = module.libname
            liburi = module.liburi
            specname = module.specname
            libpath = library_dir / libname
            syspath = module.get("syspath")
            if syspath is not None:
                syspath = library_dir / syspath
            eKonf.ensure_import_module(name, libpath, liburi, specname, syspath)



import collections
import logging
from typing import Iterator, List, Optional, Union
from tokenizers import AddedToken
from tqdm.auto import tqdm
from .base import Trainer


log = logging.getLogger(__name__)


class BpeTrainer(Trainer):
    def __init__(
        self,
        vocab_size: int = 30000,
        min_frequency: int = 2,
        special_tokens: List[Union[str, AddedToken]] = ["<unk>"],
        limit_alphabet: int = 1000,
        initial_alphabet: List[str] = [],
        end_of_word_suffix="▁",
        show_progress: bool = True,
        max_merges: int = None,
        verbose=False,
    ):
        self.vocab = {}
        self.merges = {}

        self.vocab_size = vocab_size
        self.min_frequency = min_frequency
        self.special_tokens = special_tokens
        self.limit_alphabet = limit_alphabet
        self.initial_alphabet = initial_alphabet
        self.end_of_word_suffix = end_of_word_suffix
        self.show_progress = show_progress
        self.max_merges = max_merges
        self.verbose = verbose

        super().__init__()

    def get_word_freqs(self, texts):
        word_freqs = collections.defaultdict(int)
        for text in texts:
            words = self.pre_tokenize(text)
            for word in words:
                word_freqs[word] += 1
        # Remove words that are too rare
        word_freqs = {
            word: freq
            for word, freq in word_freqs.items()
            if freq >= self.min_frequency
        }
        return word_freqs

    def initialize_vocab(self, texts):
        word_freqs = self.get_word_freqs(texts)

        alphabet = self.initial_alphabet
        for word in word_freqs.keys():
            for letter in word:
                if letter not in alphabet:
                    alphabet.append(letter)
        alphabet.sort()
        vocab = self.special_tokens + alphabet.copy() + [self.end_of_word_suffix]
        splits = {
            word: [c for c in word] + [self.end_of_word_suffix]
            for word in word_freqs.keys()
        }

        return vocab, splits, word_freqs

    def compute_pair_freqs(self, splits, word_freqs):
        pair_freqs = collections.defaultdict(int)
        for word, freq in word_freqs.items():
            split = splits[word]
            if len(split) == 1:
                continue
            for i in range(len(split) - 1):
                pair = (split[i], split[i + 1])
                pair_freqs[pair] += freq
        return pair_freqs

    def merge_pair(self, a, b, splits, word_freqs):
        for word in word_freqs:
            split = splits[word]
            if len(split) == 1:
                continue

            i = 0
            while i < len(split) - 1:
                if split[i] == a and split[i + 1] == b:
                    split = split[:i] + [a + b] + split[i + 2 :]
                else:
                    i += 1
            splits[word] = split
        return splits

    def fit(self, texts):
        vocab, splits, word_freqs = self.initialize_vocab(texts)
        vocab_size = self.vocab_size

        merges = {}
        num_merges = 0
        if self.show_progress:
            iterator = tqdm(range(vocab_size - len(vocab)))
        while len(vocab) < vocab_size:
            pair_freqs = self.compute_pair_freqs(splits, word_freqs)
            best_pair = max(pair_freqs, key=pair_freqs.get)
            splits = self.merge_pair(*best_pair, splits, word_freqs)
            merges[best_pair] = best_pair[0] + best_pair[1]
            vocab.append(best_pair[0] + best_pair[1])
            num_merges += 1
            if self.show_progress:
                iterator.update(1)
            if self.verbose:
                if (
                    isinstance(self.verbose, int) and num_merges % self.verbose == 0
                ) or (isinstance(self.verbose, list) and num_merges in self.verbose):
                    print(f"--- Round {num_merges}. Vocab size: {len(vocab)} ---")
                    print(f"Merge {num_merges}: {best_pair}")
                    print(f"Number of tokens: {len(vocab)}")
            if self.max_merges is not None and num_merges >= self.max_merges:
                break
        iterator.close()
        if self.verbose:
            print(f"--- Round {num_merges}. Vocab size: {len(vocab)} ---")
            print(f"Merge {num_merges}: {best_pair}")
            print(f"Number of tokens: {len(vocab)}")
            print(f"List of vocab: {vocab}")
        return vocab, merges

    def train(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        length: Optional[int] = None,
    ):
        """Train the model using the given iterator"""

        return self.fit(iterator)



import json
import logging
from typing import Iterator, List, Optional, Union, Dict
from ..trainers.base import Trainer


log = logging.getLogger(__name__)


class Model:
    def __init__(
        self,
        vocab=None,
        **kwargs,
    ):
        self.initialize_vocab(vocab, **kwargs)
        self.normalizer = None
        self.pre_tokenizer = None
        self.decoder = None

    def initialize_vocab(self, vocab, **kwargs):
        self.vocab = vocab
        self.token2id = {}
        self.id2token = {}
        if vocab:
            self.token2id = {token: id for id, token in enumerate(self.vocab)}
            self.id2token = {v: k for k, v in self.token2id.items()}

    def id_to_token(self, id):
        """
        Get the token associated to an ID

        Args:
            id (:obj:`int`):
                An ID to convert to a token

        Returns:
            :obj:`str`: The token associated to the ID
        """
        return self.id2token[id]

    def save(self, folder, prefix=None, pretty: bool = False):
        """
        Save the current model

        Save the current model in the given folder, using the given prefix for the various
        files that will get created.
        Any file with the same name that already exists in this folder will be overwritten.

        Args:
            folder (:obj:`str`):
                The path to the target folder in which to save the various files

            prefix (:obj:`str`, `optional`):
                An optional prefix, used to prefix each file name

        Returns:
            :obj:`List[str]`: The list of saved files
        """
        pass

    def to_str(self, pretty: bool = False):
        """Get a serialized JSON version of the Tokenizer as a str

        Args:
            pretty: bool:
                Whether the JSON string should be prettified

        Returns:
            str
        """
        return json.dumps(self.vocab, indent=2 if pretty else None)

    def token_to_id(self, token):
        """
        Get the ID associated to a token

        Args:
            token (:obj:`str`):
                A token to convert to an ID

        Returns:
            :obj:`int`: The ID associated to the token
        """
        return self.token2id[token]

    def encode(
        self,
        sequence: str,
        pair: Optional[str] = None,
        is_pretokenized: bool = False,
        add_special_tokens: bool = True,
    ):
        """Encode the given sequence and pair. This method can process raw text sequences as well
        as already pre-tokenized sequences.

        Args:
            sequence: InputSequence:
                The sequence we want to encode. This sequence can be either raw text or
                pre-tokenized, according to the `is_pretokenized` argument:

                - If `is_pretokenized=False`: `InputSequence` is expected to be `str`
                - If `is_pretokenized=True`: `InputSequence` is expected to be
                    `Union[List[str], Tuple[str]]`

            is_pretokenized: bool:
                Whether the input is already pre-tokenized.

            add_special_tokens: bool:
                Whether to add the special tokens while encoding.

        Returns:
            An Encoding
        """
        sequence_ids = [self.token_to_id(token) for token in self.tokenize(sequence)]
        # if add_special_tokens:
        #     sequence_ids = (
        #         [self.token_to_id("[CLS]")] + sequence_ids + [self.token_to_id("[SEP]")]
        #     )

        pair_ids = []
        if pair is not None:
            pair_ids = [self.token_to_id(token) for token in self.tokenize(pair)]
            # if add_special_tokens:
            #     pair_ids += [self.token_to_id("[SEP]")]

        return sequence_ids + pair_ids

    def decode(self, ids: List[int], skip_special_tokens: Optional[bool] = True) -> str:
        """Decode the given list of ids to a string sequence

        Args:
            ids: List[unsigned int]:
                A list of ids to be decoded

            skip_special_tokens: (`optional`) boolean:
                Whether to remove all the special tokens from the output string

        Returns:
            The decoded string
        """
        pass

    def tokenize(self, sequence):
        """
        Tokenize a sequence

        Args:
            sequence (:obj:`str`):
                A sequence to tokenize

        Returns:
            A :obj:`List` of :class:`~tokenizers.Token`: The generated tokens
        """
        pass

    def train(self, files: Union[str, List[str]], trainer: Trainer = None):
        """Train the model using the given files"""

        if isinstance(files, str):
            files = [files]
        texts = []
        length = 0
        for file in files:
            with open(file, "r") as f:
                texts.append(f.read())
            length += len(texts[-1])
        self.train_from_iterator(texts, trainer, length)

    def train_from_iterator(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        trainer: Trainer = None,
        length: Optional[int] = None,
    ):
        """Train the model using the given iterator"""
        pass

    def get_vocab(self, with_added_tokens: bool = True) -> Dict[str, int]:
        """Returns the vocabulary

        Args:
            with_added_tokens: boolean:
                Whether to include the added tokens in the vocabulary

        Returns:
            The vocabulary
        """
        return self.vocab

    def get_vocab_size(self, with_added_tokens: bool = True) -> int:
        """Return the size of vocabulary, with or without added tokens.

        Args:
            with_added_tokens: (`optional`) bool:
                Whether to count in added special tokens or not

        Returns:
            Size of vocabulary
        """
        return len(self.vocab)

    def normalize(self, sequence: str) -> str:
        """Normalize the given sequence

        Args:
            sequence: str:
                The sequence to normalize

        Returns:
            The normalized string
        """
        return self.normalizer.normalize_str(sequence)

    def pre_tokenize(self, sequence: str):
        """Pre-tokenize the given sequence

        Args:
            sequence: str:
                The sequence to pre-tokenize

        Returns:
            A list of tuple (str, offsets)
        """
        sequence = self.normalize(sequence)
        words_with_offsets = self.pre_tokenizer.pre_tokenize_str(sequence)
        return [word for word, _ in words_with_offsets]



import os
import logging
from typing import Iterator, List, Optional, Union
from tokenizers import (
    Regex,
    decoders,
    normalizers,
    pre_tokenizers,
)
from .models.branching import BranchingEntropy, BranchingEntropyTrainer
from .base import BaseTokenizer


log = logging.getLogger(__name__)


class BranchingEntropyTokenizer(BaseTokenizer):
    """SentencePiece Unigram Tokenizer

    Represents the Unigram algorithm, with the pretokenization used by SentencePiece
    """

    def __init__(
        self,
        vocab: Optional[str] = None,
        branching_threshold=0.2,
        whitespace_token="▁",
        whitespace_token_as_prefix=True,
        replacement: str = "▁",
        add_prefix_space: bool = True,
        lowercase: bool = True,
        verbose=False,
        **kwargs,
    ):
        if vocab is not None:
            # Let Unigram(..) fail if only one of them is None
            tokenizer = BranchingEntropy(
                vocab=vocab,
                branching_threshold=branching_threshold,
                whitespace_token=whitespace_token,
                whitespace_token_as_prefix=whitespace_token_as_prefix,
                verbose=verbose,
            )
        else:
            tokenizer = BranchingEntropy()

        normalizers_ = [
            normalizers.Nmt(),
            normalizers.NFKC(),
            normalizers.Replace(Regex(" {2,}"), " "),
        ]
        if lowercase:
            normalizers_ += [normalizers.Lowercase()]

        tokenizer.normalizer = normalizers.Sequence(normalizers_)
        tokenizer.pre_tokenizer = pre_tokenizers.Metaspace(
            replacement=replacement, add_prefix_space=add_prefix_space
        )
        tokenizer.decoder = decoders.Metaspace(
            replacement=replacement, add_prefix_space=add_prefix_space
        )

        parameters = {
            "model": "BranchingEntropy",
            "branching_threshold": branching_threshold,
            "whitespace_token": whitespace_token,
            "whitespace_token_as_prefix": whitespace_token_as_prefix,
            "replacement": replacement,
            "add_prefix_space": add_prefix_space,
        }

        super().__init__(tokenizer, parameters)

    def train(
        self,
        files: Union[str, List[str]],
        whitespace_token="▁",
        max_sentencepiece_length=20,
        min_frequency: int = 2,
        show_progress=True,
        verbose=False,
    ):
        """
        Train the model using the given files
        """

        trainer = BranchingEntropyTrainer(
            whitespace_token=whitespace_token,
            max_sentencepiece_length=max_sentencepiece_length,
            min_frequency=min_frequency,
            show_progress=show_progress,
            verbose=verbose,
        )

        if isinstance(files, str):
            files = [files]
        self._tokenizer.train(files, trainer=trainer)

    def train_from_iterator(
        self,
        iterator: Union[Iterator[str], Iterator[Iterator[str]]],
        whitespace_token="▁",
        max_sentencepiece_length=20,
        min_frequency: int = 2,
        show_progress=True,
        verbose=False,
        length: Optional[int] = None,
    ):
        """
        Train the model using the given iterator
        """

        trainer = BranchingEntropyTrainer(
            whitespace_token=whitespace_token,
            max_sentencepiece_length=max_sentencepiece_length,
            min_frequency=min_frequency,
            show_progress=show_progress,
            verbose=verbose,
        )

        self._tokenizer.train_from_iterator(
            iterator,
            trainer=trainer,
            length=length,
        )

    @staticmethod
    def load(folder, prefix=None, **kwargs):
        """Load a model from the given folder

        Args:
            folder (str): Path to the folder containing the vocab and merges files
            prefix (:obj:`str`, `optional`):
                An optional prefix, used to prefix each file name
        """
        if prefix is not None:
            folder = os.path.join(folder, prefix)
        vocab = os.path.join(folder, "vocab.json")
        config = os.path.join(folder, "config.json")

        return BranchingEntropyTokenizer.from_file(vocab, config, **kwargs)

    @staticmethod
    def from_file(vocab_filename: str, config_filename: str, **kwargs):
        vocab = BranchingEntropy.read_file(vocab_filename)
        config = BaseTokenizer.read_config(config_filename)
        config.update(kwargs)

        return BranchingEntropyTokenizer(vocab, **config)

    def find_local_entropy(self, word, direction="forward"):
        return self._tokenizer.find_local_entropy(word, direction=direction)

    def plot_local_entropy(self, word, direction="forward", figsize=(12, 5)):
        self._tokenizer.plot_local_entropy(word, direction=direction, figsize=figsize)

    def tokenize(
        self, sequence, direction="forward", flatten=True, branching_threshold=None
    ):
        return self._tokenizer.tokenize(
            sequence,
            direction=direction,
            flatten=flatten,
            branching_threshold=branching_threshold,
        )

    def naive_segment(self, text, direction="forward"):
        return self._tokenizer.naive_segment(text, direction=direction)



import os
import logging
from ekorpkit import eKonf
from .base import BaseModel
from .config import AnimMode


log = logging.getLogger(__name__)


class BatchDiffusion(BaseModel):
    def __init__(self, root_dir=None, config_name="default", **args):
        cfg = eKonf.compose(f"model/stable_diffusion={config_name}")
        cfg = eKonf.merge(cfg, args)
        super().__init__(root_dir=root_dir, **cfg)

    def batch_imagine(
        self,
        text_prompts=None,
        image_prompts=None,
        batch_name=None,
        batch_args=None,
        batch_pairs=None,
        show_collage=True,
        **args,
    ):
        """Run a batch"""
        animation_mode = AnimMode.NONE
        num_samples = 1
        args.update(dict(show_collage=False))

        if batch_args is None:
            batch_args = {}
        if not isinstance(batch_args, dict):
            raise ValueError("batch_args must be a dictionary")
        if len(batch_args) < 1:
            raise ValueError("batch_args must have at least 1 element")
        for k, v in batch_args.items():
            if not isinstance(v, (list, tuple)):
                v = [v]
        if batch_pairs is None:
            batch_pairs = [[arg_name] for arg_name in batch_args.keys()]

        _batch_results = []
        for batch_pair in batch_pairs:
            batch_pair_args = {}
            for arg_name in batch_pair:
                batch_pair_args[arg_name] = batch_args[arg_name]
            _batch_name = batch_name + "_" + "_".join(batch_pair)
            batch_config = dict(
                batch_name=_batch_name,
                batch_pair_args=batch_pair_args,
                text_prompts=text_prompts,
                image_prompts=image_prompts,
                results=[],
            )
            for pair_args in eKonf.dict_product(batch_pair_args):
                _args = args.copy()
                _args.update(pair_args)
                print(f"batch: {_batch_name} with {pair_args}")
                results = self.imagine(
                    text_prompts=text_prompts,
                    image_prompts=image_prompts,
                    batch_name=_batch_name,
                    animation_mode=animation_mode,
                    num_samples=num_samples,
                    **_args,
                )
                batch_config["results"].append(
                    dict(
                        args=pair_args,
                        config_file=results["config_file"],
                        image_filepaths=results["image_filepaths"],
                    )
                )

            _batch_config_path = self.save_batch_configs(batch_config)
            _batch_results.append(_batch_config_path)
            if show_collage:
                self.batch_collage(_batch_config_path)
        return _batch_results

    def save_batch_configs(self, args):
        """Save the settings"""
        _batch_name = args["batch_name"]
        _filename = f"{_batch_name}_batch_configs.yaml"
        _path = os.path.join(self._output.batch_configs_dir, _filename)
        log.info(f"Saving batch configs to {_path}")
        eKonf.save(args, _path)
        return _path

    def batch_collage(
        self,
        batch_config_path,
        xlabel=None,
        ylabel=None,
        zlabel=None,
        prompt_fontsize=18,
        show_filename=False,
        filename_offset=(5, 5),
        fontname=None,
        fontsize=18,
        fontcolor=None,
        **kwargs,
    ):
        args = eKonf.load(batch_config_path)
        args = eKonf.to_dict(args)

        batch_name = args["batch_name"]
        batch_pair_args = args["batch_pair_args"]
        arg_names = list(batch_pair_args.keys())

        if ylabel is None:
            ylabel = arg_names[0]
        if ylabel in arg_names:
            yticklabels = batch_pair_args[ylabel]
            arg_names.remove(ylabel)
        else:
            raise ValueError(f"{ylabel} not in {arg_names}")
        # reverse yticklabels so that the first image is the top left
        yticklabels = yticklabels[::-1]

        if xlabel is None and len(arg_names) > 0:
            xlabel = arg_names[0]
        if xlabel in arg_names:
            xticklabels = batch_pair_args[xlabel]
            arg_names.remove(xlabel)
        else:
            xticklabels = None

        if zlabel is None and len(arg_names) > 0:
            zlabel = arg_names[0]
        if zlabel in arg_names:
            ztitles = batch_pair_args[zlabel]
            arg_names.remove(zlabel)
        else:
            ztitles = [None]

        results = args["results"]
        prompt = args["text_prompts"]
        ncols = 1 if xticklabels is None else len(xticklabels)

        log.info(f"Prompt: {prompt}")
        for ztitle in ztitles:
            title = f"batch name: {batch_name}\nprompt: {prompt}\n"
            if ztitle is not None:
                output_filepath = batch_config_path.replace(
                    ".yaml", f"_{zlabel}({ztitle}).png"
                )
                title += f"{zlabel}: {ztitle}\n"
            else:
                output_filepath = batch_config_path.replace(".yaml", ".png")

            image_filepaths = []
            for result in results:
                if zlabel is None or result["args"][zlabel] == ztitle:
                    image_filepaths.append(result["image_filepaths"][0])

            eKonf.collage(
                image_filepaths=image_filepaths,
                output_filepath=output_filepath,
                ncols=ncols,
                title=title,
                title_fontsize=prompt_fontsize,
                show_filename=show_filename,
                filename_offset=filename_offset,
                fontname=fontname,
                fontsize=fontsize,
                fontcolor=fontcolor,
                xlabel=xlabel,
                ylabel=ylabel,
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                xlabel_fontsize=fontsize,
                ylabel_fontsize=fontsize,
                **kwargs,
            )
            print(f"Saved collage to {output_filepath}")



import logging
from ekorpkit import eKonf
from ekorpkit.batch import BaseConfig
from ekorpkit.visualize.collage import collage, label_collage
from .config import (
    BatchRunConfig,
    BatchConfig,
    ImagineConfig,
    RunConfig,
    CollageConfig,
    BatchRunCfg,
    BatchImagineResult,
)


log = logging.getLogger(__name__)


class BaseModel(BaseConfig):
    config_to_save = ["batch", "imagine"]

    def __init__(self, root_dir=None, **args):
        super().__init__(root_dir=root_dir, **args)
        super().__init__(**args)

        self.load_modules()
        self.sample_imagepaths = []

    def load(self):
        log.info("> downloading models...")
        self.download_models()
        log.info("> loading models...")
        self.load_models()

    def imagine(self, **args):
        """Imagine the text prompts"""
        raise NotImplementedError

    def load_models(self):
        raise NotImplementedError

    def download_models(self):
        """Download the models"""
        pass
        # download = self.args.download
        # for name, model in download.models.items():
        #     if not isinstance(model, str):
        #         log.info(f"Downloading model {name} from {model}")

    def get_run_config(self, config):
        batch = BatchConfig(output_dir=config.path.output_dir, **config.batch)
        imagine = ImagineConfig(**config.imagine)
        collage = CollageConfig(**config.collage)
        rc = RunConfig(batch=batch, imagine=imagine, collage=collage)
        return rc

    def collage(
        self,
        images_or_uris=None,
        batch_name=None,
        batch_num=None,
        display_collage=True,
        save_collage=True,
        clear_output=True,
        show_prompt=True,
        ncols=3,
        max_images=12,
        collage_width=1200,
        padding: int = 10,
        bg_color: str = "black",
        crop_to_min_size=False,
        show_filename=False,
        filename_offset=(5, 5),
        fontname=None,
        fontsize=12,
        fontcolor="#000",
        **kwargs,
    ):
        config = self.load_config(batch_name, batch_num, **kwargs)
        rc = self.get_run_config(config)

        if images_or_uris is None:
            images_or_uris = rc.image_filepaths
        collage_filepath = rc.batch.collage_filepath if save_collage else None

        collage_result = collage(
            images_or_uris=images_or_uris,
            collage_filepath=collage_filepath,
            ncols=ncols,
            max_images=max_images,
            collage_width=collage_width,
            padding=padding,
            bg_color=bg_color,
            crop_to_min_size=crop_to_min_size,
            show_filename=show_filename,
            filename_offset=filename_offset,
            fontname=fontname,
            fontsize=fontsize,
            fontcolor=fontcolor,
            **kwargs,
        )
        if display_collage and collage_result is not None:
            if clear_output:
                eKonf.clear_output(wait=True)
            if show_prompt:
                prompt = rc.imagine.get_str_prompt()
                if prompt:
                    print(f"Prompt: {prompt}")
            img = collage_result.image
            if rc.batch.max_display_image_width is not None:
                img = eKonf.scale_image(img, max_width=rc.batch.max_display_image_width)
            eKonf.display(img)

    def batch_imagine(
        self,
        batch_name,
        batch_run_params,
        batch_run_pairs=None,
        num_samples=1,
        max_display_image_width=800,
        **imagine_args,
    ):
        """Run a batch"""
        if num_samples is not None:
            imagine_args.update(dict(num_samples=num_samples))

        batch = BatchRunConfig(
            output_dir=self.output_dir,
            batch_name=batch_name,
            batch_run_params=batch_run_params,
            batch_run_pairs=batch_run_pairs,
        )

        run_configs = {}
        collage_filepaths = {}
        batch_prompts = {}
        for batch_run_cfg in batch.batch_run_configs:
            batch_run_pair = batch_run_cfg.batch_run_pair
            batch_run_name = batch_run_cfg.batch_name
            for batch_args in batch.get_run_args(batch_run_pair):
                imgn_args = batch.get_imagine_config(batch_args, **imagine_args)
                batch_prompts[batch_run_name] = None
                if "text_prompts" in imgn_args:
                    batch_args["text_prompts"] = imgn_args["text_prompts"]
                    batch_prompts[batch_run_name] = imgn_args["text_prompts"]
                log.info(f"batch: {batch_run_name} with {batch_args}")
                imagine_rst = self.imagine(
                    batch_name=batch_run_name,
                    **imgn_args,
                )
                batch_run_cfg.append(batch_args, imagine_rst)

            run_config_path = batch.save_run_config(batch_run_cfg)
            run_configs[batch_run_name] = run_config_path

        if batch.display_collage or batch.save_collage:
            eKonf.clear_output(wait=True)
            for batch_run_name, run_config_path in run_configs.items():
                collage_filepaths[batch_run_name] = self.batch_collage(
                    run_config_path, max_display_image_width=max_display_image_width
                )
        return BatchImagineResult(
            run_configs=run_configs,
            collage_filepaths=collage_filepaths,
            batch_prompts=batch_prompts,
        )

    def batch_collage(
        self,
        run_config_path,
        max_collages=5,
        display_collage=True,
        clear_output=False,
        xlabel=None,
        ylabel=None,
        zlabel=None,
        ncols=3,
        max_images_per_collage=20,
        prompt_fontsize=12,
        show_filename=False,
        filename_offset=(5, 5),
        fontname=None,
        fontsize=18,
        fontcolor=None,
        max_display_image_width=800,
        dpi=100,
        **kwargs,
    ):
        cfg = BatchRunCfg(run_config_path=run_config_path)
        batch_name = cfg.batch_name
        batch_run_pair = cfg.batch_run_pair
        arg_names = list(batch_run_pair.keys())

        if "text_prompts" in arg_names:
            zlabel = "text_prompts"
            zvalues = batch_run_pair[zlabel]
            arg_names.remove(zlabel)
        else:
            zvalues = [None]

        if ylabel is None and len(arg_names) > 0:
            ylabel = arg_names[0]
        if ylabel in arg_names:
            yticklabels = batch_run_pair[ylabel]
            # reverse yticklabels so that the first image is the top left
            yticklabels = yticklabels[::-1]
            arg_names.remove(ylabel)
        else:
            yticklabels = None
            # raise ValueError(f"{ylabel} not in {arg_names}")

        if xlabel is None and len(arg_names) > 0:
            xlabel = arg_names[0]
        if xlabel in arg_names:
            xticklabels = batch_run_pair[xlabel]
            arg_names.remove(xlabel)
        else:
            xticklabels = None

        if zlabel is None and len(arg_names) > 0:
            zlabel = arg_names[0]
        if zlabel in arg_names:
            zvalues = batch_run_pair[zlabel]
            arg_names.remove(zlabel)

        if xticklabels is None and zlabel is not None and zlabel != "text_prompts":
            ncols = 1
        elif xticklabels:
            ncols = len(xticklabels)

        collage_filepaths = []
        prompt = ""
        if len(zvalues) > max_collages:
            zvalues = zvalues[:max_collages]
        for z, zvalue in enumerate(zvalues):

            title = f"batch name: {batch_name}"
            if zvalue is not None:
                if zlabel == "text_prompts":
                    sfx = f"{zlabel}({z})"
                else:
                    sfx = f"{zlabel}({zvalue})"
                    title += f"\n{zlabel}: {zvalue}"
                output_filepath = cfg.collage_filepath(sfx)
            else:
                output_filepath = cfg.collage_filepath()
            collage_filepaths.append(output_filepath)
            image_filepaths = []
            for result in cfg.imagine_results:
                if zvalue is not None and zlabel == "text_prompts" and xlabel is None:
                    if result.batch_args[zlabel] == zvalue:
                        image_filepaths = result.image_filepaths
                        prompt = zvalue
                        break
                if zlabel is None or result.batch_args[zlabel] == zvalue:
                    image_filepaths.append(result.image_filepaths[0])
                    prompt = result.batch_args.get("text_prompts", None)

            if len(image_filepaths) == 0:
                continue
            if prompt:
                title += f"\nPrompt: {prompt}"

            collage_result = collage(
                images_or_uris=image_filepaths,
                collage_filepath=output_filepath,
                ncols=ncols,
                max_images=max_images_per_collage,
                show_filename=show_filename,
                filename_offset=filename_offset,
                fontname=fontname,
                fontsize=fontsize,
                fontcolor=fontcolor,
                **kwargs,
            )

            collage_result = label_collage(
                collage_result,
                collage_filepath=output_filepath,
                title=title,
                title_fontsize=prompt_fontsize,
                xlabel=xlabel,
                ylabel=ylabel,
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                xlabel_fontsize=fontsize,
                ylabel_fontsize=fontsize,
                dpi=dpi,
                **kwargs,
            )
            if display_collage and collage_result is not None:
                if clear_output:
                    eKonf.clear_output(wait=True)
                if prompt:
                    print(f"Prompt: {prompt}")
                img = collage_result.image
                if max_display_image_width is not None:
                    img = eKonf.scale_image(img, max_width=max_display_image_width)
                eKonf.display(img)

        return collage_filepaths

    def save_config(self, config=None):
        super().save_config(config=config, selected=self.config_to_save)
